# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from numpy.testing import assert_allclose, assert_equal
from astropy_healpix import healpy as hp
from ..testing import make_test_wcs_geometry
from ..healpix import (
    healpix_order_to_npix,
    healpix_pixel_corners,
    healpix_pixels_in_sky_image,
    hips_order_for_pixel_resolution,
    hips_tile_healpix_ipix_array,
)


def test_order_to_npix():
    assert healpix_order_to_npix(3) == 768


def test_healpix_pixel_corners():
    """
    These HEALPix corner values were verified through Aladin Lite with the "Show
    healpix grid" option turned on. More information can be found on this GitHub
    issue: https://github.com/healpy/healpy/issues/393#issuecomment-305994042
    """
    corners = healpix_pixel_corners(order=3, ipix=450, frame='icrs')

    assert_allclose(corners.ra.deg, [264.375, 258.75, 264.375, 270.])
    assert_allclose(corners.dec.deg, [-24.624318, -30., -35.685335, -30.])


@pytest.mark.parametrize('pars', [
    dict(frame='galactic', ipix=[269, 270, 271, 280, 282, 283, 292, 293, 295, 304, 305, 306]),
    dict(frame='icrs', ipix=[448, 449, 450, 451, 454, 456, 457, 460, 661, 663, 669]),
])
def test_wcs_healpix_pixel_indices(pars):
    geometry = make_test_wcs_geometry()
    healpix_pixel_indices = healpix_pixels_in_sky_image(geometry, order=3, healpix_frame=pars['frame'])
    assert list(healpix_pixel_indices) == pars['ipix']


@pytest.mark.parametrize('pars', [
    dict(tile_width=512, resolution=0.01232, resolution_res=0.06395791924665553, order=4),
    dict(tile_width=256, resolution=0.0016022, resolution_res=0.003997369952915971, order=8),
    dict(tile_width=128, resolution=0.00009032, resolution_res=0.00012491781102862408, order=13),
])
def test_get_hips_order_for_resolution(pars):
    hips_order = hips_order_for_pixel_resolution(pars['tile_width'], pars['resolution'])
    assert hips_order == pars['order']
    hips_resolution = hp.nside2resol(hp.order2nside(hips_order))
    assert_allclose(hips_resolution, pars['resolution_res'])


def test_hips_tile_healpix_ipix_array():
    ipix = hips_tile_healpix_ipix_array(shift_order=1)
    ipix_expected = [
        [0, 1],
        [2, 3],
    ]
    assert ipix.shape == (2, 2)
    assert_equal(ipix, ipix_expected)

    ipix = hips_tile_healpix_ipix_array(shift_order=2)
    ipix_expected = [
        [0, 1, 4, 5],
        [2, 3, 6, 7],
        [8, 9, 12, 13],
        [10, 11, 14, 15],
    ]
    assert ipix.shape == (4, 4)
    assert_equal(ipix, ipix_expected)

    ipix = hips_tile_healpix_ipix_array(shift_order=3)
    assert ipix.shape == (8, 8)
    assert ipix[0, 0] == 0
    assert ipix[-1, -1] == 63



"""Basic example how to plot a sky image with the hips package"""
from astropy.coordinates import SkyCoord
from hips import WCSGeometry, make_sky_image

# Compute the sky image
geometry = WCSGeometry.create(
    skydir=SkyCoord(0, 0, unit='deg', frame='galactic'),
    width=2000, height=1000, fov="3 deg",
    coordsys='galactic', projection='AIT',
)
hips_survey = 'CDS/P/DSS2/red'
result = make_sky_image(geometry=geometry, hips_survey=hips_survey, tile_format='fits')
result.plot()

# Draw the sky image
# import matplotlib.pyplot as plt
# from astropy.visualization.mpl_normalize import simple_norm
# ax = plt.subplot(projection=geometry.wcs)
# norm = simple_norm(result.image, 'sqrt', min_percent=1, max_percent=99)
# ax.imshow(result.image, origin='lower', norm=norm, cmap='gray')

# import matplotlib.pyplot as plt
# plt.show()



#!/usr/bin/env python

"""
Setuptools bootstrapping installer.

Maintained at https://github.com/pypa/setuptools/tree/bootstrap.

Run this script to install or upgrade setuptools.

This method is DEPRECATED. Check https://github.com/pypa/setuptools/issues/581 for more details.
"""

import os
import shutil
import sys
import tempfile
import zipfile
import optparse
import subprocess
import platform
import textwrap
import contextlib

from distutils import log

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

try:
    from site import USER_SITE
except ImportError:
    USER_SITE = None

# 33.1.1 is the last version that supports setuptools self upgrade/installation.
DEFAULT_VERSION = "33.1.1"
DEFAULT_URL = "https://pypi.io/packages/source/s/setuptools/"
DEFAULT_SAVE_DIR = os.curdir
DEFAULT_DEPRECATION_MESSAGE = "ez_setup.py is deprecated and when using it setuptools will be pinned to {0} since it's the last version that supports setuptools self upgrade/installation, check https://github.com/pypa/setuptools/issues/581 for more info; use pip to install setuptools"

MEANINGFUL_INVALID_ZIP_ERR_MSG = 'Maybe {0} is corrupted, delete it and try again.'

log.warn(DEFAULT_DEPRECATION_MESSAGE.format(DEFAULT_VERSION))


def _python_cmd(*args):
    """
    Execute a command.

    Return True if the command succeeded.
    """
    args = (sys.executable,) + args
    return subprocess.call(args) == 0


def _install(archive_filename, install_args=()):
    """Install Setuptools."""
    with archive_context(archive_filename):
        # installing
        log.warn('Installing Setuptools')
        if not _python_cmd('setup.py', 'install', *install_args):
            log.warn('Something went wrong during the installation.')
            log.warn('See the error message above.')
            # exitcode will be 2
            return 2


def _build_egg(egg, archive_filename, to_dir):
    """Build Setuptools egg."""
    with archive_context(archive_filename):
        # building an egg
        log.warn('Building a Setuptools egg in %s', to_dir)
        _python_cmd('setup.py', '-q', 'bdist_egg', '--dist-dir', to_dir)
    # returning the result
    log.warn(egg)
    if not os.path.exists(egg):
        raise IOError('Could not build the egg.')


class ContextualZipFile(zipfile.ZipFile):

    """Supplement ZipFile class to support context manager for Python 2.6."""

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __new__(cls, *args, **kwargs):
        """Construct a ZipFile or ContextualZipFile as appropriate."""
        if hasattr(zipfile.ZipFile, '__exit__'):
            return zipfile.ZipFile(*args, **kwargs)
        return super(ContextualZipFile, cls).__new__(cls)


@contextlib.contextmanager
def archive_context(filename):
    """
    Unzip filename to a temporary directory, set to the cwd.

    The unzipped target is cleaned up after.
    """
    tmpdir = tempfile.mkdtemp()
    log.warn('Extracting in %s', tmpdir)
    old_wd = os.getcwd()
    try:
        os.chdir(tmpdir)
        try:
            with ContextualZipFile(filename) as archive:
                archive.extractall()
        except zipfile.BadZipfile as err:
            if not err.args:
                err.args = ('', )
            err.args = err.args + (
                MEANINGFUL_INVALID_ZIP_ERR_MSG.format(filename),
            )
            raise

        # going in the directory
        subdir = os.path.join(tmpdir, os.listdir(tmpdir)[0])
        os.chdir(subdir)
        log.warn('Now working in %s', subdir)
        yield

    finally:
        os.chdir(old_wd)
        shutil.rmtree(tmpdir)


def _do_download(version, download_base, to_dir, download_delay):
    """Download Setuptools."""
    py_desig = 'py{sys.version_info[0]}.{sys.version_info[1]}'.format(sys=sys)
    tp = 'setuptools-{version}-{py_desig}.egg'
    egg = os.path.join(to_dir, tp.format(**locals()))
    if not os.path.exists(egg):
        archive = download_setuptools(version, download_base,
            to_dir, download_delay)
        _build_egg(egg, archive, to_dir)
    sys.path.insert(0, egg)

    # Remove previously-imported pkg_resources if present (see
    # https://bitbucket.org/pypa/setuptools/pull-request/7/ for details).
    if 'pkg_resources' in sys.modules:
        _unload_pkg_resources()

    import setuptools
    setuptools.bootstrap_install_from = egg


def use_setuptools(
        version=DEFAULT_VERSION, download_base=DEFAULT_URL,
        to_dir=DEFAULT_SAVE_DIR, download_delay=15):
    """
    Ensure that a setuptools version is installed.

    Return None. Raise SystemExit if the requested version
    or later cannot be installed.
    """
    to_dir = os.path.abspath(to_dir)

    # prior to importing, capture the module state for
    # representative modules.
    rep_modules = 'pkg_resources', 'setuptools'
    imported = set(sys.modules).intersection(rep_modules)

    try:
        import pkg_resources
        pkg_resources.require("setuptools>=" + version)
        # a suitable version is already installed
        return
    except ImportError:
        # pkg_resources not available; setuptools is not installed; download
        pass
    except pkg_resources.DistributionNotFound:
        # no version of setuptools was found; allow download
        pass
    except pkg_resources.VersionConflict as VC_err:
        if imported:
            _conflict_bail(VC_err, version)

        # otherwise, unload pkg_resources to allow the downloaded version to
        #  take precedence.
        del pkg_resources
        _unload_pkg_resources()

    return _do_download(version, download_base, to_dir, download_delay)


def _conflict_bail(VC_err, version):
    """
    Setuptools was imported prior to invocation, so it is
    unsafe to unload it. Bail out.
    """
    conflict_tmpl = textwrap.dedent("""
        The required version of setuptools (>={version}) is not available,
        and can't be installed while this script is running. Please
        install a more recent version first, using
        'easy_install -U setuptools'.

        (Currently using {VC_err.args[0]!r})
        """)
    msg = conflict_tmpl.format(**locals())
    sys.stderr.write(msg)
    sys.exit(2)


def _unload_pkg_resources():
    sys.meta_path = [
        importer
        for importer in sys.meta_path
        if importer.__class__.__module__ != 'pkg_resources.extern'
    ]
    del_modules = [
        name for name in sys.modules
        if name.startswith('pkg_resources')
    ]
    for mod_name in del_modules:
        del sys.modules[mod_name]


def _clean_check(cmd, target):
    """
    Run the command to download target.

    If the command fails, clean up before re-raising the error.
    """
    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError:
        if os.access(target, os.F_OK):
            os.unlink(target)
        raise


def download_file_powershell(url, target):
    """
    Download the file at url to target using Powershell.

    Powershell will validate trust.
    Raise an exception if the command cannot complete.
    """
    target = os.path.abspath(target)
    ps_cmd = (
        "[System.Net.WebRequest]::DefaultWebProxy.Credentials = "
        "[System.Net.CredentialCache]::DefaultCredentials; "
        '(new-object System.Net.WebClient).DownloadFile("%(url)s", "%(target)s")'
        % locals()
    )
    cmd = [
        'powershell',
        '-Command',
        ps_cmd,
    ]
    _clean_check(cmd, target)


def has_powershell():
    """Determine if Powershell is available."""
    if platform.system() != 'Windows':
        return False
    cmd = ['powershell', '-Command', 'echo test']
    with open(os.path.devnull, 'wb') as devnull:
        try:
            subprocess.check_call(cmd, stdout=devnull, stderr=devnull)
        except Exception:
            return False
    return True
download_file_powershell.viable = has_powershell


def download_file_curl(url, target):
    cmd = ['curl', url, '--location', '--silent', '--output', target]
    _clean_check(cmd, target)


def has_curl():
    cmd = ['curl', '--version']
    with open(os.path.devnull, 'wb') as devnull:
        try:
            subprocess.check_call(cmd, stdout=devnull, stderr=devnull)
        except Exception:
            return False
    return True
download_file_curl.viable = has_curl


def download_file_wget(url, target):
    cmd = ['wget', url, '--quiet', '--output-document', target]
    _clean_check(cmd, target)


def has_wget():
    cmd = ['wget', '--version']
    with open(os.path.devnull, 'wb') as devnull:
        try:
            subprocess.check_call(cmd, stdout=devnull, stderr=devnull)
        except Exception:
            return False
    return True
download_file_wget.viable = has_wget


def download_file_insecure(url, target):
    """Use Python to download the file, without connection authentication."""
    src = urlopen(url)
    try:
        # Read all the data in one block.
        data = src.read()
    finally:
        src.close()

    # Write all the data in one block to avoid creating a partial file.
    with open(target, "wb") as dst:
        dst.write(data)
download_file_insecure.viable = lambda: True


def get_best_downloader():
    downloaders = (
        download_file_powershell,
        download_file_curl,
        download_file_wget,
        download_file_insecure,
    )
    viable_downloaders = (dl for dl in downloaders if dl.viable())
    return next(viable_downloaders, None)


def download_setuptools(
        version=DEFAULT_VERSION, download_base=DEFAULT_URL,
        to_dir=DEFAULT_SAVE_DIR, delay=15,
        downloader_factory=get_best_downloader):
    """
    Download setuptools from a specified location and return its filename.

    `version` should be a valid setuptools version number that is available
    as an sdist for download under the `download_base` URL (which should end
    with a '/'). `to_dir` is the directory where the egg will be downloaded.
    `delay` is the number of seconds to pause before an actual download
    attempt.

    ``downloader_factory`` should be a function taking no arguments and
    returning a function for downloading a URL to a target.
    """
    # making sure we use the absolute path
    to_dir = os.path.abspath(to_dir)
    zip_name = "setuptools-%s.zip" % version
    url = download_base + zip_name
    saveto = os.path.join(to_dir, zip_name)
    if not os.path.exists(saveto):  # Avoid repeated downloads
        log.warn("Downloading %s", url)
        downloader = downloader_factory()
        downloader(url, saveto)
    return os.path.realpath(saveto)


def _build_install_args(options):
    """
    Build the arguments to 'python setup.py install' on the setuptools package.

    Returns list of command line arguments.
    """
    return ['--user'] if options.user_install else []


def _parse_args():
    """Parse the command line for options."""
    parser = optparse.OptionParser()
    parser.add_option(
        '--user', dest='user_install', action='store_true', default=False,
        help='install in user site package')
    parser.add_option(
        '--download-base', dest='download_base', metavar="URL",
        default=DEFAULT_URL,
        help='alternative URL from where to download the setuptools package')
    parser.add_option(
        '--insecure', dest='downloader_factory', action='store_const',
        const=lambda: download_file_insecure, default=get_best_downloader,
        help='Use internal, non-validating downloader'
    )
    parser.add_option(
        '--version', help="Specify which version to download",
        default=DEFAULT_VERSION,
    )
    parser.add_option(
        '--to-dir',
        help="Directory to save (and re-use) package",
        default=DEFAULT_SAVE_DIR,
    )
    options, args = parser.parse_args()
    # positional arguments are ignored
    return options


def _download_args(options):
    """Return args for download_setuptools function from cmdline args."""
    return dict(
        version=options.version,
        download_base=options.download_base,
        downloader_factory=options.downloader_factory,
        to_dir=options.to_dir,
    )


def main():
    """Install or upgrade setuptools and EasyInstall."""
    options = _parse_args()
    archive = download_setuptools(**_download_args(options))
    return _install(archive, _build_install_args(options))

if __name__ == '__main__':
    sys.exit(main())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from typing import List, Tuple
from copy import deepcopy
import warnings
import urllib.request
from io import BytesIO
from pathlib import Path
import numpy as np
from PIL import Image
from astropy.utils.exceptions import AstropyWarning
from astropy.io.fits.verify import VerifyWarning
from astropy.coordinates import SkyCoord
from astropy.io import fits
from ..utils.healpix import healpix_pixel_corners
from .io import tile_default_url, tile_default_path

__all__ = [
    'HipsTileMeta',
    'HipsTile',
]

__doctest_skip__ = [
    'HipsTile',
]


# TODO: this could be a dict. Would that be better?
def compute_image_shape(width: int, height: int, fmt: str) -> tuple:
    """Compute numpy array shape for a given image.

    The output image shape is 2-dim for grayscale, and 3-dim for color images:

    * ``shape = (height, width)`` for FITS images with one grayscale channel
    * ``shape = (height, width, 3)`` for JPG images with three RGB channels
    * ``shape = (height, width, 4)`` for PNG images with four RGBA channels

    Parameters
    ----------
    width, height : int
        Width and height of the image
    fmt : {'fits', 'jpg', 'png'}
        Image format

    Returns
    -------
    shape : tuple
        Numpy array shape
    """
    if fmt == 'fits':
        return height, width
    elif fmt == 'jpg':
        return height, width, 3
    elif fmt == 'png':
        return height, width, 4
    else:
        raise ValueError(f'Invalid format: {fmt}')


class HipsTileMeta:
    """HiPS tile metadata.

    Parameters
    ----------
    order : int
        HEALPix order
    ipix : int
        HEALPix pixel number
    file_format : {'fits', 'jpg', 'png'}
        File format
    frame : {'icrs', 'galactic', 'ecliptic'}
        Sky coordinate frame
    width : int
        Tile width (tiles always have width = height)

    Examples
    --------
    >>> from hips import HipsTileMeta
    >>> tile_meta = HipsTileMeta(order=3, ipix=450, file_format='fits')
    >>> tile_meta
    HipsTileMeta(order=3, ipix=450, file_format='fits', frame='icrs', width=512)
    >>> tile_meta.skycoord_corners
    <SkyCoord (ICRS): (ra, dec) in deg
    [( 264.375, -24.62431835), ( 258.75 , -30.        ),
    ( 264.375, -35.68533471), ( 270.   , -30.        )]>
    >>> tile_meta.tile_default_url
    'Norder3/Dir0/Npix450.fits'
    >>> tile_meta.tile_default_path.as_posix()
    'Norder3/Dir0/Npix450.fits'
    """

    def __init__(self, order: int, ipix: int, file_format: str,
                 frame: str = 'icrs', width: int = 512) -> None:
        self.order = order
        self.ipix = ipix
        self.file_format = file_format
        self.frame = frame
        self.width = width

    def __repr__(self):
        return (
            'HipsTileMeta('
            f'order={self.order}, ipix={self.ipix}, '
            f'file_format={self.file_format!r}, frame={self.frame!r}, '
            f'width={self.width}'
            ')'
        )

    def __eq__(self, other: 'HipsTileMeta') -> bool:
        return (
            self.order == other.order and
            self.ipix == other.ipix and
            self.file_format == other.file_format and
            self.frame == other.frame and
            self.width == other.width
        )

    def copy(self):
        """An independent copy."""
        return deepcopy(self)

    @property
    def skycoord_corners(self) -> SkyCoord:
        """Tile corner sky coordinates (`~astropy.coordinates.SkyCoord`)."""
        return healpix_pixel_corners(self.order, self.ipix, self.frame)

    @property
    def tile_default_url(self) -> str:
        """Tile relative URL (str)."""
        return tile_default_url(self.order, self.ipix, self.file_format)

    @property
    def tile_default_path(self) -> Path:
        """Tile relative filename path (`~pathlib.Path`)."""
        return tile_default_path(self.order, self.ipix, self.file_format)


class HipsTile:
    """HiPS tile container.

    This class provides methods for fetching, reading, and writing a HiPS tile.

    .. note::

        In HiPS, the pixel data is flipped in the y direction for
        ``jpg`` and ``png`` format with respect to FITS.
        In this package, we handle this by flipping ``jpg`` and ``png``
        data to match the ``fits`` orientation, at the I/O boundary,
        i.e. in `from_numpy` and `to_numpy`.

    Parameters
    ----------
    meta : `~hips.HipsTileMeta`
        Metadata of HiPS tile
    raw_data : `bytes`
        Raw data (copy of bytes from file)

    Examples
    --------

    Fetch a HiPS tile:

    >>> from hips import HipsTile, HipsTileMeta
    >>> meta = HipsTileMeta(order=6, ipix=30889, file_format='fits')
    >>> url = 'http://alasky.unistra.fr/2MASS/H/Norder6/Dir30000/Npix30889.fits'
    >>> tile = HipsTile.fetch(meta, url)

    The tile pixel data is available as a Numpy array:

    >>> type(tile.data)
    numpy.ndarray
    >>> tile.data.shape
    (512, 512)
    >>> tile.data.dtype.name
    int16
    """

    def __init__(self, meta: HipsTileMeta, raw_data: bytes) -> None:
        self.meta = meta
        self.raw_data = raw_data
        self._data = None

    def __eq__(self, other: 'HipsTile') -> bool:
        return (
            self.meta == other.meta and
            self.raw_data == other.raw_data
        )

    @classmethod
    def from_numpy(cls, meta: HipsTileMeta, data: np.ndarray) -> 'HipsTile':
        """Create a tile from given pixel data.

        Parameters
        ----------
        meta : `~hips.HipsTileMeta`
            Metadata of HiPS tile
        data : `~numpy.ndarray`
            Tile pixel data

        Returns
        -------
        tile : `~hips.HipsTile`
            HiPS tile object in the format requested in ``meta``.
        """
        fmt = meta.file_format
        bio = BytesIO()

        if fmt == 'fits':
            hdu = fits.PrimaryHDU(data)
            hdu.writeto(bio)
        elif fmt == 'jpg':
            # Flip tile to be consistent with FITS orientation
            data = np.flipud(data)
            image = Image.fromarray(data)
            image.save(bio, format='jpeg')
        elif fmt == 'png':
            # Flip tile to be consistent with FITS orientation
            data = np.flipud(data)
            image = Image.fromarray(data)
            image.save(bio, format='png')
        else:
            raise ValueError(f'Tile file format not supported: {fmt}. '
                             'Supported formats: fits, jpg, png')

        bio.seek(0)
        raw_data = bio.read()

        return cls(meta, raw_data)

    @property
    def children(self) -> List['HipsTile']:
        """Create four children tiles from parent tile."""
        w = self.data.shape[0] // 2
        data = [
            self.data[w: w * 2, 0: w],
            self.data[0: w, 0: w],
            self.data[w: w * 2, w: w * 2],
            self.data[0: w, w: w * 2]
        ]

        tiles = []
        for idx in range(4):
            meta = HipsTileMeta(
                self.meta.order + 1,
                self.meta.ipix * 4 + idx,
                self.meta.file_format,
                self.meta.frame,
                len(data[0])
            )
            tile = self.from_numpy(meta, data[idx])
            tiles.append(tile)

        return tiles

    @property
    def data(self) -> np.ndarray:
        """Tile pixel data (`~numpy.ndarray`).

        This is a cached property, it will only be computed once.

        See the `to_numpy` function.
        """
        if self._data is None:
            self._data = self.to_numpy(
                self.raw_data,
                self.meta.file_format,
            )

        return self._data

    @staticmethod
    def to_numpy(raw_data: bytes, fmt: str) -> np.ndarray:
        """Convert raw image bytes to Numpy array pixel data.

        Parameters
        ----------
        raw_data : bytes
            Raw image bytes (usually read from file or fetched from URL)
        fmt : {'fits', 'jpg', 'png'}
            File format

        Returns
        -------
        data : `~numpy.ndarray`
            Pixel data as a numpy array
        """
        bio = BytesIO(raw_data)

        if fmt == 'fits':
            # At the moment CDS is serving FITS tiles in non-standard FITS files
            # https://github.com/hipspy/hips/issues/42
            # The following warnings handling is supposed to suppress these warnings
            # (but hopefully still surface other issues in a useful way).
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', AstropyWarning)
                warnings.simplefilter('ignore', VerifyWarning)
                with fits.open(bio) as hdu_list:
                    data = hdu_list[0].data
        elif fmt in {'jpg', 'png'}:
            with Image.open(bio) as image:
                data = np.array(image)
            # Flip tile to be consistent with FITS orientation
            data = np.flipud(data)
        else:
            raise ValueError(f'Tile file format not supported: {fmt}. '
                             'Supported formats: fits, jpg, png')

        return data

    @classmethod
    def read(cls, meta: HipsTileMeta, filename: str = None) -> 'HipsTile':
        """Read HiPS tile data from a directory and load into memory (`~hips.HipsTile`).

        Parameters
        ----------
        meta : `~hips.HipsTileMeta`
            Metadata of HiPS tile
        filename : str
            Filename
        """
        raw_data = Path(filename).read_bytes()
        return cls(meta, raw_data)

    @classmethod
    def fetch(cls, meta: HipsTileMeta, url: str) -> 'HipsTile':
        """Fetch HiPS tile and load into memory (`~hips.HipsTile`).

        Parameters
        ----------
        meta : `~hips.HipsTileMeta`
            Metadata of HiPS tile
        url : str
            URL containing HiPS tile
        """
        with urllib.request.urlopen(url) as response:
            raw_data = response.read()

        return cls(meta, raw_data)

    def write(self, filename: str) -> None:
        """Write to file.

        Parameters
        ----------
        filename : str
            Filename
        """
        Path(filename).write_bytes(self.raw_data)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from pathlib import Path
import pytest
from astropy.tests.helper import remote_data
from numpy.testing import assert_equal, assert_allclose
from ...utils.testing import get_hips_extra_file, requires_hips_extra
from ..tile import HipsTileMeta
from ..allsky import HipsTileAllskyArray

TEST_CASES = [
    dict(
        label='fits',
        meta=dict(order=3, ipix=-1, file_format='fits'),
        url='https://github.com/hipspy/hips-extra/blob/master/datasets/samples/IRAC4/Norder3/Allsky.fits?raw=true',
        filename='datasets/samples/IRAC4/Norder3/Allsky.fits',

        repr="HipsTileAllskyArray(format='fits', order=3, width=1728, "
             "height=1856, n_tiles=768, tile_width=64)",
        dtype='float32',
        shape=(1856, 1728),
        # This is the pixel with the maximum value, to make it easy to find
        pix_idx=([1198], [742]),
        pix_val=2174.190673828125,

        tile_shape=(64, 64),
        # This is the same pixel as above, with the max value in the all-sky image
        tile_idx=281,
        tile_pix_idx=[[46], [38]],
        tile_pix_val=2174.190673828125,

    ),
    dict(
        label='jpg-grayscale',
        meta=dict(order=3, ipix=-1, file_format='jpg'),
        url='http://alasky.unistra.fr/IRAC4/Norder3/Allsky.jpg',
        filename='datasets/samples/IRAC4/Norder3/Allsky.jpg',

        repr="HipsTileAllskyArray(format='jpg', order=3, width=1728, "
             "height=1856, n_tiles=768, tile_width=64)",
        dtype='uint8',
        shape=(1856, 1728),
        # This is the same pixel as used above in the FITS test case
        # For JPEG it's saturated at value 255 and not the unique max
        pix_idx=([1198], [742]),
        pix_val=255,

        tile_shape=(64, 64),
        # This is the same pixel as above, with the max value in the all-sky image
        tile_idx=281,
        tile_pix_idx=[[46], [38]],
        tile_pix_val=244,
    ),
    dict(
        label='jpg-rgb',
        meta=dict(order=3, ipix=-1, file_format='jpg'),
        url='http://alasky.unistra.fr/Fermi/Color/Norder3/Allsky.jpg',
        filename='datasets/samples/FermiColor/Norder3/Allsky.jpg',

        repr="HipsTileAllskyArray(format='jpg', order=3, width=1728, "
             "height=1856, n_tiles=768, tile_width=64)",
        dtype='uint8',
        shape=(1856, 1728, 3),
        pix_idx=([510], [5]),
        pix_val=[[34, 23, 5]],

        tile_shape=(64, 64, 3),
        tile_idx=0,
        tile_pix_idx=([0], [0]),
        tile_pix_val=[[62, 44, 30]],
    ),
]


def parametrize(labels=None):
    if labels:
        test_cases = [_ for _ in TEST_CASES if _['label'] in labels]
    else:
        test_cases = TEST_CASES
    ids = lambda _: _['label']
    return pytest.mark.parametrize('pars', test_cases, ids=ids)


def _read_tile(pars):
    meta = HipsTileMeta(**pars['meta'])
    filename = get_hips_extra_file(pars['filename'])
    return HipsTileAllskyArray.read(meta, filename)


@requires_hips_extra()
@parametrize()
def test_read(pars):
    # Check that reading tiles in various formats works,
    # i.e. that pixel data in numpy array format
    # has correct shape, dtype and values
    allsky = _read_tile(pars)

    assert repr(allsky) == pars['repr']
    assert allsky.meta.order == pars['meta']['order']
    assert isinstance(allsky.raw_data, bytes)

    data = allsky.data
    assert data.shape == pars['shape']
    assert data.dtype.name == pars['dtype']
    assert_allclose(allsky.data[pars['pix_idx']], pars['pix_val'])


@requires_hips_extra()
# JPEG encoding is lossy, so here we only run the FITS and PNG test cases
@parametrize(labels='fits')
def test_from_numpy(pars):
    tile = _read_tile(pars)
    tile2 = HipsTileAllskyArray.from_numpy(tile.meta, tile.data)
    assert_equal(tile.data, tile2.data)


@remote_data
@requires_hips_extra()
@parametrize()
def test_fetch(pars):
    # Check that tile HTTP fetch gives the same result as tile read from disk
    meta = HipsTileMeta(**pars['meta'])
    tile_fetch = HipsTileAllskyArray.fetch(meta, url=pars['url'])

    filename = get_hips_extra_file(pars['filename'])
    tile_read = HipsTileAllskyArray.read(meta, filename)

    assert tile_fetch == tile_read


@requires_hips_extra()
@parametrize()
def test_write(tmpdir, pars):
    # Check that tile I/O works, i.e. round-trips on write / read
    tile = _read_tile(pars)

    filename = str(tmpdir / Path(pars['filename']).name)
    tile.write(filename)
    tile2 = HipsTileAllskyArray.read(tile.meta, filename)

    assert tile == tile2


@requires_hips_extra()
@parametrize()
def test_tile(pars):
    allsky = _read_tile(pars)

    tile = allsky.tile(pars['tile_idx'])

    assert tile.data.shape == pars['tile_shape']
    assert_allclose(tile.data[pars['tile_pix_idx']], pars['tile_pix_val'])


@requires_hips_extra()
@parametrize(labels='fits')
def test_from_tiles(pars):
    # Check that ``from_tiles`` works properly
    # For now, we check that ``tiles`` and ``from_tiles`` round-trip
    # TODO: it would probably be better to test them separately,
    # asserting on each of the two step. Round-trip can work, if
    # the same mistake is made in each conversion step.
    allsky = _read_tile(pars)
    tiles = allsky.tiles

    # This is for debugging ...
    allsky.write('/tmp/allsky.fits')
    allsky2 = HipsTileAllskyArray.from_tiles(tiles)
    allsky2.write('/tmp/allsky2.fits')

    # TODO: at the moment `HipsTileAllskyArray` always does a JPG encoding,
    # it can't hold the numpy array data unchanged.
    # This still doesn't work, because when going ``allsky.tiles`` another
    # JPG encoding happens.
    data2 = HipsTileAllskyArray.tiles_to_allsky_array(tiles)
    assert_allclose(allsky.data, data2)



from setuptools.command.install import install as SetuptoolsInstall

from ..utils import _get_platlib_dir


class AstropyInstall(SetuptoolsInstall):
    user_options = SetuptoolsInstall.user_options[:]
    boolean_options = SetuptoolsInstall.boolean_options[:]

    def finalize_options(self):
        build_cmd = self.get_finalized_command('build')
        platlib_dir = _get_platlib_dir(build_cmd)
        self.build_lib = platlib_dir
        SetuptoolsInstall.finalize_options(self)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from numpy.testing import assert_allclose
from astropy.coordinates import SkyCoord
from astropy.tests.helper import remote_data
from ...utils.testing import requires_hips_extra
from ...utils.wcs import WCSGeometry
from ...tiles import HipsSurveyProperties
from ..paint import is_tile_distorted, measure_tile_lengths, HipsPainter, plot_mpl_single_tile


@remote_data
class TestHipsPainter:
    @classmethod
    def setup_class(cls):
        url = 'http://alasky.unistra.fr/DSS/DSS2Merged/properties'
        cls.hips_survey = HipsSurveyProperties.fetch(url)
        cls.geometry = WCSGeometry.create(
            skydir=SkyCoord(0, 0, unit='deg', frame='icrs'),
            width=2000, height=1000, fov="3 deg",
            coordsys='icrs', projection='AIT',
        )
        fetch_opts = dict(fetch_package='urllib', timeout=30, n_parallel=10)
        cls.painter = HipsPainter(cls.geometry, cls.hips_survey, 'fits', fetch_opts=fetch_opts)

    def test_draw_hips_order(self):
        assert self.painter.draw_hips_order == 7

    def test_tile_indices(self):
        assert list(self.painter.tile_indices)[:4] == [69623, 69627, 69628, 69629]

    draw_hips_order_pars = [
        dict(order=7, fov="3 deg"),
        dict(order=5, fov="10 deg"),
        dict(order=4, fov="15 deg"),
    ]

    @requires_hips_extra()
    @pytest.mark.parametrize('pars', draw_hips_order_pars)
    def test_compute_matching_hips_order(self, pars):
        geometry = WCSGeometry.create(
            skydir=SkyCoord(0, 0, unit='deg', frame='icrs'),
            width=2000, height=1000, fov=pars['fov'],
            coordsys='icrs', projection='AIT',
        )

        fetch_opts = dict(fetch_package='urllib', timeout=30, n_parallel=10)
        simple_tile_painter = HipsPainter(geometry, self.hips_survey, 'fits', fetch_opts=fetch_opts)
        assert simple_tile_painter.draw_hips_order == pars['order']

    def test_run(self):
        self.painter.run()
        assert self.painter.image.shape == (1000, 2000)
        assert_allclose(self.painter.image[200, 994], 2120)

    def test_draw_hips_tile_grid(self):
        self.painter.plot_mpl_hips_tile_grid()

    def test_draw_debug_image(self):
        tile = self.painter.tiles[3]
        image = self.painter._make_empty_sky_image()
        plot_mpl_single_tile(self.geometry, tile, image)


@pytest.fixture(scope='session')
def corners():
    x = [764.627476, 999., 764.646551, 530.26981]
    y = [300.055412, 101.107245, -97.849955, 101.105373]
    return x, y


def test_is_tile_distorted(corners):
    assert is_tile_distorted(corners) is True


def test_measure_tile_lengths(corners):
    edges, diagonals = measure_tile_lengths(corners)

    expected = [307.426175, 307.417479, 307.434024, 307.41606]
    assert_allclose(edges, expected)

    assert_allclose(diagonals, [397.905367, 468.73019])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Tile input / output (I/O) methods (both local and remote)."""
from typing import List
from pathlib import Path

__all__ = [
    'tile_default_url',
    'tile_default_path',
]


def _tile_default_location(order: int, ipix: int, file_format: str) -> List[str]:
    # Directory index: HiPS tiles are grouped in chunks of 10k tiles
    dir_idx = (ipix // 10_000) * 10_000

    return [
        f'Norder{order}',
        f'Dir{dir_idx}',
        f'Npix{ipix}.{file_format}',
    ]


def tile_default_url(order: int, ipix: int, file_format: str) -> str:
    loc = _tile_default_location(order, ipix, file_format)
    return '/'.join(loc)


def tile_default_path(order: int, ipix: int, file_format: str) -> Path:
    loc = _tile_default_location(order, ipix, file_format)
    return Path(*loc)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A Python astronomy package for HiPS : Hierarchical Progressive Surveys.

At the moment a client for HiPS images, but other contributions
(HiPS catalogs or HiPS image generation) welcome!

* Code : https://github.com/hipspy/hips
* Docs : https://hips.readthedocs.io
* License : BSD-3 (see licenses folder for license file)
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *
# ----------------------------------------------------------------------------

if not _ASTROPY_SETUP_:
    # For egg_info test builds to pass, put package imports here.
    from .draw import *
    from .tiles import *
    from .utils.wcs import WCSGeometry



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import time
import numpy as np
from typing import List, Tuple, Union, Dict, Any
from astropy.wcs.utils import proj_plane_pixel_scales
from skimage.transform import ProjectiveTransform, warp
from ..tiles import HipsSurveyProperties, HipsTile, HipsTileMeta, fetch_tiles
from ..tiles.tile import compute_image_shape
from ..utils.wcs import WCSGeometry
from ..utils.healpix import healpix_pixels_in_sky_image, hips_order_for_pixel_resolution

__all__ = [
    'HipsPainter',
]

__doctest_skip__ = [
    'HipsPainter',
]


class HipsPainter:
    """Paint a sky image from HiPS image tiles.

    Paint HiPS tiles onto a ky image using a simple projective
    transformation method. The algorithm implemented is described
    here: :ref:`drawing_algo`.

    Parameters
    ----------
    geometry : dict or `~hips.utils.WCSGeometry`
        An object of WCSGeometry
    hips_survey : str or `~hips.HipsSurveyProperties`
        HiPS survey properties
    tile_format : {'fits', 'jpg', 'png'}
        Format of HiPS tile
    precise : bool
        Use the precise drawing algorithm
    progress_bar : bool
        Show a progress bar for tile fetching and drawing
    fetch_opts : dict
        Keyword arguments for fetching HiPS tiles. To see the
        list of passable arguments, refer to `~fetch_tiles`

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> from hips import WCSGeometry
    >>> from hips import HipsSurveyProperties
    >>> from hips import HipsPainter
    >>> geometry = WCSGeometry.create(
    ...     skydir=SkyCoord(0, 0, unit='deg', frame='icrs'),
    ...     width=2000, height=1000, fov='3 deg',
    ...     coordsys='icrs', projection='AIT',
    ... )
    >>> url = 'http://alasky.unistra.fr/DSS/DSS2Merged/properties'
    >>> hips_survey = HipsSurveyProperties.fetch(url)
    >>> painter = HipsPainter(geometry, hips_survey, 'fits')
    >>> painter.draw_hips_order
    7
    >>> painter.run()
    >>> painter.image.shape
    (1000, 2000)
    """

    def __init__(self, geometry: Union[dict, WCSGeometry], hips_survey: Union[str, HipsSurveyProperties],
                 tile_format: str, precise: bool = False, progress_bar: bool = True, fetch_opts : dict = None) -> None:
        self.geometry = WCSGeometry.make(geometry)
        self.hips_survey = HipsSurveyProperties.make(hips_survey)
        self.tile_format = tile_format
        self.precise = precise
        self.progress_bar = progress_bar
        self.fetch_opts = fetch_opts
        self._tiles = None
        self.float_image = None
        self._stats: Dict[str, Any] = {}

    @property
    def image(self) -> np.ndarray:
        """Computed sky image (`~numpy.ndarray`).

        * The ``dtype`` is always chosen to match the tile ``dtype``.
          This is ``uint8`` for JPG or PNG tiles,
          and can be e.g. ``int16`` or ``float32`` for FITS tiles.
        * The output shape is documented here: `~HipsPainter.shape`.
        """
        return self.float_image.astype(self.tiles[0].data.dtype)

    @property
    def draw_hips_order(self) -> int:
        """Compute HiPS tile order matching a given image pixel size."""
        # Sky image angular resolution (pixel size in degrees)
        resolution = np.min(proj_plane_pixel_scales(self.geometry.wcs))
        desired_order = hips_order_for_pixel_resolution(self.hips_survey.tile_width, resolution)
        # Return the desired order, or the highest resolution available.
        # Note that HiPS never has resolution less than 3,
        # and that limit is handled in _get_hips_order_for_resolution
        return np.min([desired_order, self.hips_survey.hips_order])

    @property
    def tile_indices(self):
        """Get list of index values for HiPS tiles."""
        return healpix_pixels_in_sky_image(
            geometry=self.geometry,
            order=self.draw_hips_order,
            healpix_frame=self.hips_survey.astropy_frame,
        )

    def projection(self, tile: HipsTile) -> ProjectiveTransform:
        """Estimate projective transformation on a HiPS tile."""
        corners = tile.meta.skycoord_corners.to_pixel(self.geometry.wcs)
        src = np.array(corners).T.reshape((4, 2))
        dst = tile_corner_pixel_coordinates(tile.meta.width)
        pt = ProjectiveTransform()
        pt.estimate(src, dst)
        return pt

    @property
    def tiles(self) -> List[HipsTile]:
        """List of `~hips.HipsTile` (cached on multiple access)."""
        tile_metas = []
        for healpix_pixel_index in self.tile_indices:
            tile_meta = HipsTileMeta(
                order=self.draw_hips_order,
                ipix=healpix_pixel_index,
                frame=self.hips_survey.astropy_frame,
                file_format=self.tile_format,
            )
            tile_metas.append(tile_meta)

        if self._tiles is None:
            self._tiles = fetch_tiles(tile_metas=tile_metas, hips_survey=self.hips_survey,
                                      progress_bar=self.progress_bar, **(self.fetch_opts or {}))

        return self._tiles

    def warp_image(self, tile: HipsTile) -> np.ndarray:
        """Warp a HiPS tile and a sky image."""
        return warp(
            tile.data,
            self.projection(tile),
            output_shape=self.geometry.shape,
            preserve_range=True,
        )

    def run(self) -> np.ndarray:
        """Draw HiPS tiles onto an empty image."""
        t0 = time.time()
        self.make_tile_list()
        t1 = time.time()
        self._stats['fetch_time'] = t1 - t0

        t0 = time.time()
        self.draw_all_tiles()
        t1 = time.time()
        self._stats['draw_time'] = t1 - t0

        self._stats['tile_count'] = len(self.draw_tiles)
        self._stats['consumed_memory'] = 0
        for tile in self.draw_tiles:
            self._stats['consumed_memory'] += len(tile.raw_data)


    def make_tile_list(self):
        parent_tiles = self.tiles

        # Suggestion: for now, just split once, no recursion.
        # Leave TODO, to discuss with Thomas next week.
        # See also: https://github.com/hipspy/hips/issues/92

        if self.precise == True:
            tiles = []
            for tile in parent_tiles:
                corners = tile.meta.skycoord_corners.to_pixel(self.geometry.wcs)
                if is_tile_distorted(corners):
                    tiles.append(tile.children)
                else:
                    tiles.append(tile)
            self.draw_tiles = [tile for children in tiles for tile in children]
        else:
            self.draw_tiles = parent_tiles

    def _make_empty_sky_image(self):
        shape = compute_image_shape(
            width=self.geometry.shape.width,
            height=self.geometry.shape.height,
            fmt=self.tile_format,
        )
        return np.zeros(shape, dtype=np.float32)

    def draw_all_tiles(self):
        """Make an empty sky image and draw all the tiles."""
        image = self._make_empty_sky_image()
        if self.progress_bar:
            from tqdm import tqdm
            tiles = tqdm(self.draw_tiles, desc='Drawing tiles')
        else:
            tiles = self.draw_tiles

        for tile in tiles:
            tile_image = self.warp_image(tile)
            # TODO: put better algorithm here instead of summing pixels
            # this can lead to pixels that are painted twice and become to bright
            image += tile_image

        # Store the result
        self.float_image = image

    def plot_mpl_hips_tile_grid(self) -> None:
        """Plot output image and HiPS grid with matplotlib.

        This is mainly useful for debugging the drawing algorithm,
        not something end-users will call or need to know about.
        """
        import matplotlib.pyplot as plt
        self.make_tile_list()
        for tile in self.draw_tiles:
            corners = tile.meta.skycoord_corners.transform_to(self.geometry.celestial_frame)
            ax = plt.subplot(projection=self.geometry.wcs)
            opts = dict(color='red', lw=1, )
            ax.plot(corners.data.lon.deg, corners.data.lat.deg,
                    transform=ax.get_transform('world'), **opts)
        ax.imshow(self.image, origin='lower')


def measure_tile_lengths(corners: Tuple[np.ndarray, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
    """Compute length of tile edges and diagonals.

    Parameters
    ----------
    corners : tuple of `~numpy.ndarray`
        Tile corner pixel coordinates ``(x, y)``

    Returns
    -------
    edges : `~numpy.ndarray`
        Tile edge pixel lengths.
        Entries: 0 -> 1, 1 -> 2, 2 -> 3, 3 -> 0
    diagonals : `~numpy.ndarray`
        Tile diagonal pixel lengths
        Entries: 0 -> 2, 1 -> 3
    """
    x, y = corners

    def dist(i: int, j: int) -> float:
        """Compute distance between two points."""
        return np.sqrt((x[i] - x[j]) ** 2 + (y[i] - y[j]) ** 2)

    edges = [dist((i + 1) % 4, i) for i in range(4)]
    diagonals = [dist(0, 2), dist(1, 3)]

    return np.array(edges), np.array(diagonals)


def is_tile_distorted(corners: tuple) -> bool:
    """Is the tile with the given corners distorted?

    The criterion implemented here is described on the
    :ref:`drawing_algo` page, as part of the tile drawing algorithm.
    """
    edges, diagonals = measure_tile_lengths(corners)
    diagonal_ratio = min(diagonals) / max(diagonals)

    return bool(
        max(edges) > 300 or
        (max(diagonals) > 150 and
        diagonal_ratio < 0.7)
    )


def tile_corner_pixel_coordinates(width) -> np.ndarray:
    """Tile corner pixel coordinates for projective transform.

    Note that in this package we flip pixel data from JPEG and PNG
    to be consistent with FITS tile orientation on read.
    For that reason we can treat all tiles the same here and in other
    places in the package, and assume the FITS tile orientation is given.

    The order of corners below is chosen such that it matches the order
    of the pixel corner sky coordinates from ``healpix_pixel_corners``:

    - north
    - west
    - south
    - east

    and then gives correct results when used to compute the projective transform for tile drawing.
    """
    w = width - 1
    return np.array([
        [w, 0],  # north
        [w, w],  # west
        [0, w],  # south
        [0, 0],  # east
    ])


def plot_mpl_single_tile(geometry: WCSGeometry, tile: HipsTile, image: np.ndarray) -> None:
    """Draw markers on the output image (mainly used for debugging).

    The following denotes their correspondence:
    * red <=> North
    * green <=> West
    * blue <=> South
    * yellow <=> East

    Parameters
    ----------
    geometry : `~hips.utils.WCSGeometry`
        Geometry of the output image
    tile : HipsTile
        HiPS tile
    image : np.ndarray
        Image containing HiPS tiles
    """
    import matplotlib.pyplot as plt
    corners = tile.meta.skycoord_corners.transform_to(geometry.celestial_frame)
    colors = ['red', 'green', 'blue', 'yellow']
    ax = plt.subplot(projection=geometry.wcs)
    for index, corner in enumerate(corners):
        opts = dict(s=80, color=colors[index])
        ax.scatter(corner.data.lon.deg, corner.data.lat.deg,
                   transform=ax.get_transform('world'), **opts)
    ax.imshow(image, origin='lower')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from astropy.tests.helper import remote_data
from numpy.testing import assert_allclose
from astropy.coordinates import SkyCoord
from ..wcs import WCSGeometry


class TestWCSGeometry:
    def setup(self):
        self.geometry = WCSGeometry.create(
            skydir=SkyCoord(0, 0, unit='deg', frame='galactic'),
            width=2000, height=1000, fov='3 deg',
            coordsys='galactic', projection='AIT',
        )

    def test_galactic_frame_center(self):
        c = self.geometry.center_skycoord

        assert c.frame.name == 'galactic'
        assert_allclose(c.l.deg, 359.99, atol=1e-2)
        assert_allclose(c.b.deg, 0.00075, atol=1e-2)
        assert_allclose(self.geometry.wcs.wcs.crpix, [1000., 500.])
        assert_allclose(self.geometry.wcs.wcs.cdelt, [-0.0015, 0.0015])

    def test_celestial_frame(self):
        geometry = WCSGeometry.create(
            skydir=SkyCoord(0, 0, unit='deg', frame='icrs'),
            width=2000, height=1000, fov='3 deg',
            coordsys='icrs', projection='AIT',
        )
        c = geometry.center_skycoord

        assert c.frame.name == 'icrs'
        assert_allclose(c.ra.deg, 359.99, atol=1e-2)
        assert_allclose(c.dec.deg, 0.00075, atol=1e-2)
        assert_allclose(self.geometry.wcs.wcs.crpix, [1000., 500.])
        assert_allclose(self.geometry.wcs.wcs.cdelt, [-0.0015, 0.0015])

    @remote_data
    def test_create_from_dict(self):
        params = dict(target='crab', width=2000, height=1000, fov='3 deg',
                      coordsys='galactic', projection='AIT')
        geometry = WCSGeometry.create_from_dict(params)

        c = geometry.center_skycoord
        assert c.frame.name == 'galactic'
        assert_allclose(c.l.deg, 184.55, atol=1e-2)
        assert_allclose(c.b.deg, -5.78, atol=1e-2)
        assert_allclose(self.geometry.wcs.wcs.crpix, [1000., 500.])
        assert_allclose(self.geometry.wcs.wcs.cdelt, [-0.0015, 0.0015])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""The high-level end user interface (UI)."""
import numpy as np
from PIL import Image
from astropy.io import fits
from typing import List, Union
from ..utils.wcs import WCSGeometry
from ..tiles import HipsSurveyProperties, HipsTile
from .paint import HipsPainter

__all__ = [
    'make_sky_image',
    'HipsDrawResult',
]


def make_sky_image(geometry: Union[dict, WCSGeometry], hips_survey: Union[str, 'HipsSurveyProperties'],
                   tile_format: str, precise: bool = False, progress_bar: bool = True, fetch_opts: dict = None) -> 'HipsDrawResult':
    """Make sky image: fetch tiles and draw.

    The example for this can be found on the :ref:`gs` page.

    Parameters
    ----------
    geometry : dict or `~hips.utils.WCSGeometry`
        Geometry of the output image
    hips_survey : str or `~hips.HipsSurveyProperties`
        HiPS survey properties
    tile_format : {'fits', 'jpg', 'png'}
        Format of HiPS tile to use
        (some surveys are available in several formats, so this extra argument is needed)
    precise : bool
        Use the precise drawing algorithm
    progress_bar : bool
        Show a progress bar for tile fetching and drawing
    fetch_opts : dict
        Keyword arguments for fetching HiPS tiles. To see the
        list of passable arguments, refer to `~hips.fetch_tiles`

    Returns
    -------
    result : `~hips.HipsDrawResult`
        Result object
    """
    painter = HipsPainter(geometry, hips_survey, tile_format, precise, progress_bar, fetch_opts)
    painter.run()
    return HipsDrawResult.from_painter(painter)


class HipsDrawResult:
    """HiPS draw result object (sky image and more).

    Parameters
    ----------
    image : `~numpy.ndarray`
        Sky image (the main result)
    geometry : `~hips.utils.WCSGeometry`
        WCS geometry of the sky image
    tile_format : {'fits', 'jpg', 'png'}
        Format of HiPS tile
    tiles : list
        Python list of `~hips.HipsTile` objects that were used
    stats : dict
        Information including time for fetching and drawing HiPS tiles
    """

    def __init__(self, image: np.ndarray, geometry: WCSGeometry, tile_format: str, tiles: List[HipsTile], stats: dict) -> None:
        self.image = image
        self.geometry = geometry
        self.tile_format = tile_format
        self.tiles = tiles
        self.stats = stats

    def __str__(self):
        return (
            'HiPS draw result:\n'
            f'Sky image: shape={self.image.shape}, dtype={self.image.dtype}\n'
            f'WCS geometry: {self.geometry}\n'
        )

    def __repr__(self):
        return (
            'HipsDrawResult('
            f'width={self.image.shape[0]}, '
            f'height={self.image.shape[1]}, '
            f'channels={self.image.ndim}, '
            f'dtype={self.image.dtype}, '
            f'format={self.tile_format}'
            ')'
        )

    @classmethod
    def from_painter(cls, painter: HipsPainter) -> 'HipsDrawResult':
        """Make a `~hips.HipsDrawResult` from a `~hips.HipsTilePainter`."""
        return cls(
            image=painter.image,
            geometry=painter.geometry,
            tile_format=painter.tile_format,
            tiles=painter.tiles,
            stats=painter._stats,
        )

    def write_image(self, filename: str) -> None:
        """Write image to file.

        Parameters
        ----------
        filename : str
            Filename
        """
        if self.tile_format == 'fits':
            hdu = fits.PrimaryHDU(data=self.image, header=self.geometry.fits_header)
            hdu.writeto(filename)
        else:
            image = Image.fromarray(self.image)
            image.save(filename)

    def plot(self) -> None:
        """Plot the all sky image and overlay HiPS tile outlines.

        Uses `astropy.visualization.wcsaxes`.
        """
        import matplotlib.pyplot as plt
        for tile in self.tiles:
            corners = tile.meta.skycoord_corners
            corners = corners.transform_to(self.geometry.celestial_frame)
            ax = plt.subplot(projection=self.geometry.wcs)
            opts = dict(color='red', lw=1, )
            ax.plot(corners.data.lon.deg, corners.data.lat.deg,
                    transform=ax.get_transform('world'), **opts)
        ax.imshow(self.image, origin='lower')

    def report(self) -> None:
        """Print a brief report for the fetched data."""

        print (
            f"Time for fetching tiles = {self.stats['fetch_time']} seconds\n"
            f"Time for drawing tiles = {self.stats['draw_time']} seconds\n"
            f"Total memory consumed = {self.stats['consumed_memory'] / 1e6} MB\n"
            f"Total tiles fetched = {self.stats['tile_count']}\n"
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from numpy.testing import assert_allclose
from astropy.utils.data import get_pkg_data_filename
from astropy.tests.helper import remote_data
from ...utils.testing import get_hips_extra_file, requires_hips_extra
from ..tile import HipsTileMeta
from ..survey import HipsSurveyProperties, HipsSurveyPropertiesList


class TestHipsSurveyProperties:
    @classmethod
    def setup_class(cls):
        filename = get_pkg_data_filename('data/properties.txt')
        cls.survey = HipsSurveyProperties.read(filename)

    @remote_data
    def test_from_name(self):
        survey = HipsSurveyProperties.from_name('CDS/P/2MASS/color')
        assert survey.title == '2MASS color J (1.23um), H (1.66um), K (2.16um)'

    @remote_data
    def test_make(self):
        survey = HipsSurveyProperties.make('CDS/P/EGRET/Dif/300-500')
        assert survey.title == 'EGRET Dif 300-500MeV'
        assert self.survey is HipsSurveyProperties.make(self.survey)

    def test_title(self):
        assert self.survey.title == 'DSS colored'

    def test_hips_version(self):
        assert self.survey.hips_version == '1.31'

    def test_hips_frame(self):
        assert self.survey.hips_frame == 'equatorial'

    def test_astropy_frame(self):
        assert self.survey.astropy_frame == 'icrs'

    def test_hips_order(self):
        assert self.survey.hips_order == 9

    def test_tile_format(self):
        assert self.survey.tile_format == 'jpeg'

    def test_base_url(self):
        expected = 'http://alasky.u-strasbg.fr/DSS/DSSColor'
        assert self.survey.base_url == expected

    def test_tile_default_url(self):
        tile_meta = HipsTileMeta(order=9, ipix=54321, file_format='fits')
        url = self.survey.tile_url(tile_meta)
        assert url == 'http://alasky.u-strasbg.fr/DSS/DSSColor/Norder9/Dir50000/Npix54321.fits'

    @staticmethod
    @requires_hips_extra()
    def test_tile_width():
        filename = get_hips_extra_file('datasets/samples/Planck-HFI143/properties')
        survey = HipsSurveyProperties.read(filename)
        assert survey.tile_width == 256

    @staticmethod
    @remote_data
    def test_fetch():
        url = 'http://alasky.u-strasbg.fr/DSS/DSS2-NIR/properties'
        survey = HipsSurveyProperties.fetch(url)
        assert survey.base_url == 'http://alasky.u-strasbg.fr/DSS/DSS2-NIR'


class TestHipsSurveyPropertiesList:
    @classmethod
    def setup_class(cls):
        filename = get_pkg_data_filename('data/surveys.txt')
        cls.surveys = HipsSurveyPropertiesList.read(filename)

    def test_surveys(self):
        assert len(self.surveys.data) == 4

        survey = self.surveys.data[0]
        assert survey.data['creator_did'] == 'ivo://CDS/C/MUSE-M42'
        assert survey.data['obs_collection'] == 'MUSE-M42'
        assert survey.data['hips_tile_format'] == 'png fits'

    def test_table(self):
        table = self.surveys.table
        assert len(table) == 4

        row = table[0]
        assert row['creator_did'] == 'ivo://CDS/C/MUSE-M42'
        assert row['obs_collection'] == 'MUSE-M42'
        assert row['hips_tile_format'] == 'png fits'
        assert row['moc_order'] == 12
        assert_allclose(row['moc_sky_fraction'], 2.98e-07, rtol=0.01)

    @remote_data
    def test_fetch(self):
        surveys = HipsSurveyPropertiesList.fetch()
        assert len(surveys.data) > 3

        survey = surveys.from_name('CDS/P/2MASS/H')
        assert survey.title == '2MASS H (1.66um)'
        assert survey.hips_order == 9

    @remote_data
    def test_key_error(self):
        with pytest.raises(KeyError):
            surveys = HipsSurveyPropertiesList.fetch()
            surveys.from_name('Kronka Lonka')



"""Basic example how to plot a sky image with the hips package"""
from astropy.coordinates import SkyCoord
from hips import WCSGeometry, make_sky_image

# Compute the sky image
geometry = WCSGeometry.create(
    skydir=SkyCoord(0, 0, unit='deg', frame='galactic'),
    width=2000, height=1000, fov="3 deg",
    coordsys='galactic', projection='AIT',
)
hips_survey = 'IPAC/P/GLIMPSE360'
result = make_sky_image(geometry=geometry, hips_survey=hips_survey, tile_format='jpg')

# Draw the sky image
import matplotlib.pyplot as plt
ax = plt.subplot(projection=geometry.wcs)
ax.imshow(result.image, origin='lower')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from typing import List
import numpy as np
from ..utils.healpix import healpix_order_to_npix
from .tile import HipsTile

__all__ = [
    'HipsTileAllskyArray',
]

__doctest_skip__ = [
    'HipsTileAllskyArray',
]


class HipsTileAllskyArray(HipsTile):
    """All-sky tile array container.

    To quote from section 4.3.2 "Allsky preview file" of the HiPS IVOA working draft:
    "The tiles at low orders (0 to 3) may be packaged
    together into a unique file called Allsky."

    This class implements that all-sky tile array format.

    TODO: We're sub-classing `~hips.HipsTile` here at the moment.
    This is weird!
    Probably the current ``HipsTile`` should be renamed ``ImageIO``
    or be split up into functions that do image I/O in the three supported formats?

    TODO: We re-use the `~hips.HipsTileMeta` class to store ``order``
    as well as other info like ``file_format`` and ``frame``.
    Note that ``ipix`` doesn't make sense for an ``AllSkyTileArray``.
    Maybe there's  a better way to handle this without code duplication?

    Examples
    --------

    Load an example existing HiPS all-sky image
    (unfortunately one has to pass a dummy ipix value here):

    >>> from hips import HipsTileAllskyArray, HipsTileMeta
    >>> meta = HipsTileMeta(order=3, ipix=-1, file_format='jpg', frame='icrs')
    >>> url = 'http://alasky.unistra.fr/Fermi/Color/Norder3/Allsky.jpg'
    >>> allsky = HipsTileAllskyArray.fetch(meta, url)

    Now you can extract tiles (e.g. for drawing):

    >>> tile = allsky.tile(ipix=42)
    >>> tile.meta
    HipsTileMeta(order=3, ipix=42, file_format='jpg', frame='icrs', width=64)

    TODO: add an example how to go the other way: combine tiles into an allsky image.
    """

    def __repr__(self):
        return (
            'HipsTileAllskyArray('
            f'format={self.meta.file_format!r}, order={self.meta.order}, '
            f'width={self.width}, height={self.height}, '
            f'n_tiles={self.n_tiles}, '
            f'tile_width={self.tile_width}'
            ')'
        )

    @property
    def width(self) -> int:
        """Image pixel width (int)."""
        return self.data.shape[1]

    @property
    def height(self) -> int:
        """Image pixel height (int)"""
        return self.data.shape[0]

    @property
    def n_tiles(self) -> int:
        """Number of tiles in the image (int)."""
        return healpix_order_to_npix(self.meta.order)

    @property
    def n_tiles_in_row(self) -> int:
        """Number of tiles per tile row (int)."""
        return int(np.sqrt(self.n_tiles))

    @property
    def tile_width(self) -> int:
        """Pixel width of a single tile (int)."""
        return self.width // self.n_tiles_in_row

    @classmethod
    def from_tiles(cls, tiles: List[HipsTile]) -> 'HipsTileAllskyArray':
        """Create all-sky image from list of tiles."""
        meta = tiles[0].meta.copy()
        data = cls.tiles_to_allsky_array(tiles)
        # TODO: check return type here.
        # Pycharm warns that a `HipsTile` is returned here, not a `HipsTileAllskyArray`
        # Is this true or a bug in their static code analysis?
        return cls.from_numpy(meta, data)

    @staticmethod
    def tiles_to_allsky_array(tiles: List[HipsTile]) -> np.ndarray:
        """Combine tiles into an all-sky image."""
        # Compute all-sky image parameters that we need below
        n_tiles = len(tiles)
        n_tiles_in_row = int(np.sqrt(n_tiles))
        n_tiles_in_col = (n_tiles // n_tiles_in_row) + 1
        tile_width = tiles[0].meta.width

        # Make an empty all-sky image
        shape = (
            tile_width * n_tiles_in_col,  # height
            tile_width * n_tiles_in_row,  # width
        )
        if len(tiles[0].data.shape) == 3:
            shape = (*shape, tiles[0].data.shape[2])

        # TODO: is this a good way to fill the blank pixels?
        # I checked `datasets/samples/DSS2Red/Norder3/Allsky.fits` and
        # there in the FITS header and data array I see
        # BLANK = -32768.0
        # I don't do this here for now because I don't know how to
        # handle this properly for FITS / JPEG / PNG
        # Maybe we should switch to `NaN` or `-32768` for float data?
        blank_value = 0
        data = blank_value * np.ones(shape, tiles[0].data.dtype)

        # Copy over the tile data into the all-sky image
        for tile in tiles:
            tile_slice = HipsTileAllskyArray._tile_slice(
                ipix=tile.meta.ipix,
                tile_width=tile.meta.width,
                n_tiles_in_row=n_tiles_in_row,
            )
            # TODO: explain this flip. (see comment below in the tile method)
            data[::-1, ...][tile_slice] = tile.data[::-1, ...]

        return data

    @property
    def tiles(self) -> List[HipsTile]:
        """Split into a list of `~hips.HipsTile`.

        This is called when using the all-sky image for drawing.
        """
        return [self.tile(ipix) for ipix in range(self.n_tiles)]

    def tile(self, ipix: int) -> HipsTile:
        """Extract one of the tiles (`~hips.HipsTile`)

        A copy of the data by default.
        For drawing we could avoid the copy by passing ``copy=False`` here.
        """
        meta = self.meta.copy()
        meta.ipix = ipix
        meta.width = self.tile_width

        tile_slice = self._tile_slice(ipix, self.tile_width, self.n_tiles_in_row)
        # Note: apparently, tiles in all-sky files are ordered top to bottom,
        # left to right, always assuming the JPEG / PNG orientation, even for
        # FITS all-sky images!?
        # That's why we flip the all-sky image to JPEG / PNG orientation here
        # before extracting the tile data, and then we flip the tile data back
        # because internally we're always using the FITS tile orientation
        data = self.data[::-1, ...][tile_slice][::-1, ...]

        # data = self.data[tile_slice]

        return HipsTile.from_numpy(meta, data)

    @staticmethod
    def _tile_slice(ipix, tile_width, n_tiles_in_row):
        """Compute the 2-dim slice in the allsky ``data`` for a given tile."""
        w = tile_width
        row_idx, col_idx = divmod(ipix, n_tiles_in_row)
        return [
            slice(row_idx * w, (row_idx + 1) * w),
            slice(col_idx * w, (col_idx + 1) * w),
        ]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utilities for HiPS package testing.

Not of use for users / outside this package.
"""
import os
from pathlib import Path
import pytest
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from .wcs import WCSGeometry


def get_hips_extra_file(filename):
    """Get `~pathlib.Path` for a file in ``hips-extra``.

    To make this work, clone this repo:
    https://github.com/hipspy/hips-extra
    and set an environment variable ``HIPS_EXTRA`` pointing to it.
    """
    path = Path(os.environ['HIPS_EXTRA'])
    return path / filename


def has_hips_extra():
    """Is hips-extra available? (bool)"""
    if 'HIPS_EXTRA' in os.environ:
        path = Path(os.environ['HIPS_EXTRA']) / 'datasets/samples/DSS2Red/properties'
        if path.is_file():
            return True
    return False


def requires_hips_extra():
    """Decorator to mark tests requiring ``hips-extra`` data."""
    skip_it = not has_hips_extra()
    reason = 'No hips-extra data available.'
    return pytest.mark.skipif(skip_it, reason=reason)


def make_test_wcs_geometry():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype[0] = 'GLON-AIT'
    wcs.wcs.ctype[1] = 'GLAT-AIT'
    wcs.wcs.crval[0] = 0
    wcs.wcs.crval[1] = 0
    wcs.wcs.crpix[0] = 1000
    wcs.wcs.crpix[1] = 500
    wcs.wcs.cdelt[0] = -0.01
    wcs.wcs.cdelt[1] = 0.01
    return WCSGeometry(wcs, width=2000, height=1000)

    # geometry = WCSGeometry.create(
    #     skydir=SkyCoord(0, 0, unit='deg', frame='galactic'),
    #     width=2000, height=1000, coordsys='galactic',
    #     projection='AIT', cdelt=0.01, crpix=(1000, 500),
    # )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from pathlib import Path
import pytest
from astropy.tests.helper import remote_data
from numpy.testing import assert_allclose, assert_equal
from ...utils.testing import get_hips_extra_file, requires_hips_extra
from ..tile import HipsTileMeta, HipsTile


class TestHipsTileMeta:
    @classmethod
    def setup_class(cls):
        # Note: we're intentionally using positional args
        # to make sure tests fail if the order of args is changed
        cls.meta = HipsTileMeta(3, 450, 'fits', 'icrs', 512)

    def test_order(self):
        assert self.meta.order == 3

    def test_ipix(self):
        assert self.meta.ipix == 450

    def test_file_format(self):
        assert self.meta.file_format == 'fits'

    def test_frame(self):
        assert self.meta.frame == 'icrs'

    def test_width(self):
        assert self.meta.width == 512

    def test_repr(self):
        expected = "HipsTileMeta(order=3, ipix=450, file_format='fits', frame='icrs', width=512)"
        assert repr(self.meta) == expected

    def test_eq(self):
        assert self.meta == HipsTileMeta(3, 450, 'fits', 'icrs', 512)
        assert self.meta != HipsTileMeta(4, 450, 'fits', 'icrs', 512)

    def test_copy(self):
        meta = self.meta.copy()
        meta.ipix = 42
        assert self.meta.ipix == 450
        assert meta.ipix == 42

    def test_skycoord_corners(self):
        coord = self.meta.skycoord_corners
        assert_allclose(coord.data.lat.deg, [-24.624318, -30., -35.685335, -30.])
        assert_allclose(coord.data.lon.deg, [264.375, 258.75, 264.375, 270.])
        assert self.meta.skycoord_corners.frame.name == 'icrs'

    @staticmethod
    def test_skycoord_corners_galactic():
        meta = HipsTileMeta(order=3, ipix=450, file_format='fits', frame='galactic')
        coord = meta.skycoord_corners
        assert_allclose(coord.data.lat.deg, [-24.624318, -30., -35.685335, -30.])
        assert_allclose(coord.data.lon.deg, [264.375, 258.75, 264.375, 270.])
        assert coord.frame.name == 'galactic'

    def test_tile_url(self):
        url = self.meta.tile_default_url
        assert url == 'Norder3/Dir0/Npix450.fits'

    def test_tile_path(self):
        path = self.meta.tile_default_path
        assert path == Path('Norder3/Dir0/Npix450.fits')


HIPS_TILE_TEST_CASES = [
    dict(
        label='fits',
        meta=dict(order=3, ipix=463, file_format='fits'),
        url='http://alasky.unistra.fr/DSS/DSS2Merged/Norder3/Dir0/Npix463.fits',
        filename='datasets/samples/DSS2Red/Norder3/Dir0/Npix463.fits',

        dtype='int16',
        shape=(512, 512),
        pix_idx=[[510], [5]],
        pix_val=[3047],

        child_order=4,
        child_shape=(256, 256),
        child_ipix=[1852, 1853, 1854, 1855],
        child_pix_idx=[[255], [255]],
        child_pix_val=[2437, 2153, 2124, 2418],
    ),
    dict(
        label='jpg',
        meta=dict(order=3, ipix=463, file_format='jpg'),
        url='http://alasky.unistra.fr/Fermi/Color/Norder3/Dir0/Npix451.jpg',
        filename='datasets/samples/FermiColor/Norder3/Dir0/Npix451.jpg',

        dtype='uint8',
        shape=(512, 512, 3),
        pix_idx=[[510], [5]],
        pix_val=[[132, 116, 83]],

        child_order=4,
        child_shape=(256, 256, 3),
        child_ipix=[1852, 1853, 1854, 1855],
        child_pix_idx=[[255], [255]],
        child_pix_val=[[[255, 241, 225]], [[245, 214, 211]], [[109, 95, 86]], [[137, 97, 87]]],
    ),
    dict(
        label='png',
        meta=dict(order=6, ipix=6112, file_format='png'),
        url='http://alasky.unistra.fr/2MASS6X/2MASS6X_H/Norder6/Dir0/Npix6112.png',
        filename='datasets/samples/2MASS6XH/Norder6/Dir0/Npix6112.png',

        dtype='uint8',
        shape=(512, 512, 4),
        pix_idx=[[253], [5]],
        pix_val=[[15, 15, 15, 255]],

        child_order=7,
        child_shape=(256, 256, 4),
        child_ipix=[24448, 24449, 24450, 24451],
        child_pix_idx=[[255], [255]],
        child_pix_val=[[[15,  15,  15, 255]], [[17, 17, 17, 255]], [[20, 20, 20, 255]], [[13, 13, 13, 255]]],
    ),
]


def parametrize():
    ids = lambda _: _['label']
    return pytest.mark.parametrize('pars', HIPS_TILE_TEST_CASES, ids=ids)


class TestHipsTile:
    @staticmethod
    def _read_tile(pars):
        meta = HipsTileMeta(**pars['meta'])
        filename = get_hips_extra_file(pars['filename'])
        return HipsTile.read(meta, filename)

    @requires_hips_extra()
    @parametrize()
    def test_from_numpy(self, pars):
        tile = self._read_tile(pars)
        tile2 = HipsTile.from_numpy(tile.meta, tile.data)

        # JPEG encoding is lossy. So in that case, output pixel value
        # aren't exactly the same as input pixel values
        if tile.meta.file_format != 'jpg':
            assert_equal(tile.data, tile2.data)

    @requires_hips_extra()
    @parametrize()
    def test_read(self, pars):
        # Check that reading tiles in various formats works,
        # i.e. that pixel data in numpy array format
        # has correct shape, dtype and values
        tile = self._read_tile(pars)

        assert tile.meta.order == pars['meta']['order']
        assert isinstance(tile.raw_data, bytes)

        data = tile.data
        assert data.shape == pars['shape']
        assert data.dtype.name == pars['dtype']
        assert_equal(data[pars['pix_idx']], pars['pix_val'])

    @remote_data
    @requires_hips_extra()
    @parametrize()
    def test_fetch(self, pars):
        # Check that tile HTTP fetch gives the same result as tile read from disk
        meta = HipsTileMeta(**pars['meta'])
        tile_fetch = HipsTile.fetch(meta, url=pars['url'])

        filename = get_hips_extra_file(pars['filename'])
        tile_read = HipsTile.read(meta, filename)

        assert tile_fetch == tile_read

    @requires_hips_extra()
    @parametrize()
    def test_write(self, tmpdir, pars):
        # Check that tile I/O works, i.e. round-trips on write / read
        tile = self._read_tile(pars)

        filename = str(tmpdir / Path(pars['filename']).name)
        tile.write(filename)
        tile2 = HipsTile.read(tile.meta, filename)

        assert tile == tile2

    @requires_hips_extra()
    @parametrize()
    def test_children(self, pars):
        tile = self._read_tile(pars)
        child_data = [_.data[pars['child_pix_idx']] for _ in tile.children]
        child_ipix = [_.meta.ipix for _ in tile.children]

        assert tile.children[0].meta.order == pars['child_order']
        assert tile.children[0].data.shape == pars['child_shape']
        assert_equal(child_ipix, pars['child_ipix'])
        assert_equal(child_data, pars['child_pix_val'])



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Astropy shared Sphinx settings.  These settings are shared between
# astropy itself and affiliated packages.
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys
import warnings

from os import path

import sphinx
from distutils.version import LooseVersion


# -- General configuration ----------------------------------------------------

# The version check in Sphinx itself can only compare the major and
# minor parts of the version number, not the micro.  To do a more
# specific version check, call check_sphinx_version("x.y.z.") from
# your project's conf.py
needs_sphinx = '1.3'


on_rtd = os.environ.get('READTHEDOCS', None) == 'True'


def check_sphinx_version(expected_version):
    sphinx_version = LooseVersion(sphinx.__version__)
    expected_version = LooseVersion(expected_version)
    if sphinx_version < expected_version:
        raise RuntimeError(
            "At least Sphinx version {0} is required to build this "
            "documentation.  Found {1}.".format(
                expected_version, sphinx_version))


# Configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/',
               (None, 'http://data.astropy.org/intersphinx/python3.inv')),
    'pythonloc': ('http://docs.python.org/',
                  path.abspath(path.join(path.dirname(__file__),
                                         'local/python3_local_links.inv'))),
    'numpy': ('https://docs.scipy.org/doc/numpy/',
              (None, 'http://data.astropy.org/intersphinx/numpy.inv')),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/',
              (None, 'http://data.astropy.org/intersphinx/scipy.inv')),
    'matplotlib': ('http://matplotlib.org/',
                   (None, 'http://data.astropy.org/intersphinx/matplotlib.inv')),
    'astropy': ('http://docs.astropy.org/en/stable/', None),
    'h5py': ('http://docs.h5py.org/en/latest/', None)}

if sys.version_info[0] == 2:
    intersphinx_mapping['python'] = (
        'https://docs.python.org/2/',
        (None, 'http://data.astropy.org/intersphinx/python2.inv'))
    intersphinx_mapping['pythonloc'] = (
        'http://docs.python.org/',
        path.abspath(path.join(path.dirname(__file__),
                               'local/python2_local_links.inv')))

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# The reST default role (used for this markup: `text`) to use for all
# documents. Set to the "smart" one.
default_role = 'obj'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog = """
.. _Astropy: http://astropy.org
"""

# A list of warning types to suppress arbitrary warning messages. We mean to
# override directives in astropy_helpers.sphinx.ext.autodoc_enhancements,
# thus need to ignore those warning. This can be removed once the patch gets
# released in upstream Sphinx (https://github.com/sphinx-doc/sphinx/pull/1843).
# Suppress the warnings requires Sphinx v1.4.2
suppress_warnings = ['app.add_directive', ]

# -- Project information ------------------------------------------------------

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
#pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []


# -- Settings for extensions and extension options ----------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'astropy_helpers.extern.numpydoc',
    'astropy_helpers.extern.automodapi.automodapi',
    'astropy_helpers.extern.automodapi.smart_resolver',
    'astropy_helpers.sphinx.ext.tocdepthfix',
    'astropy_helpers.sphinx.ext.doctest',
    'astropy_helpers.sphinx.ext.changelog_links']


if not on_rtd and LooseVersion(sphinx.__version__) < LooseVersion('1.4'):
    extensions.append('sphinx.ext.pngmath')
else:
    extensions.append('sphinx.ext.mathjax')

try:
    import matplotlib.sphinxext.plot_directive
    extensions += [matplotlib.sphinxext.plot_directive.__name__]
# AttributeError is checked here in case matplotlib is installed but
# Sphinx isn't.  Note that this module is imported by the config file
# generator, even if we're not building the docs.
except (ImportError, AttributeError):
    warnings.warn(
        "matplotlib's plot_directive could not be imported. " +
        "Inline plots will not be included in the output")

# Don't show summaries of the members in each class along with the
# class' docstring
numpydoc_show_class_members = False

autosummary_generate = True

automodapi_toctreedirnm = 'api'

# Class documentation should contain *both* the class docstring and
# the __init__ docstring
autoclass_content = "both"

# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    '-Nfontsize=10',
    '-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Efontsize=10',
    '-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif',
    '-Gfontsize=10',
    '-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif'
]

# -- Options for HTML output -------------------------------------------------

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [path.abspath(path.join(path.dirname(__file__), 'themes'))]

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'bootstrap-astropy'

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    '**': ['localtoc.html'],
    'search': [],
    'genindex': [],
    'py-modindex': [],
}

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.

# included in the bootstrap-astropy theme
html_favicon = path.join(html_theme_path[0], html_theme, 'static',
                         'astropy_logo.ico')

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%d %b %Y'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# -- Options for LaTeX output ------------------------------------------------

# The paper size ('letter' or 'a4').
#latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
#latex_font_size = '10pt'

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
latex_toplevel_sectioning = 'part'

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

latex_elements = {}

# Additional stuff for the LaTeX preamble.
latex_elements['preamble'] = r"""
% Use a more modern-looking monospace font
\usepackage{inconsolata}

% The enumitem package provides unlimited nesting of lists and enums.
% Sphinx may use this in the future, in which case this can be removed.
% See https://bitbucket.org/birkenfeld/sphinx/issue/777/latex-output-too-deeply-nested
\usepackage{enumitem}
\setlistdepth{15}

% In the parameters section, place a newline after the Parameters
% header.  (This is stolen directly from Numpy's conf.py, since it
% affects Numpy-style docstrings).
\usepackage{expdlist}
\let\latexdescription=\description
\def\description{\latexdescription{}{} \breaklabel}

% Support the superscript Unicode numbers used by the "unicode" units
% formatter
\DeclareUnicodeCharacter{2070}{\ensuremath{^0}}
\DeclareUnicodeCharacter{00B9}{\ensuremath{^1}}
\DeclareUnicodeCharacter{00B2}{\ensuremath{^2}}
\DeclareUnicodeCharacter{00B3}{\ensuremath{^3}}
\DeclareUnicodeCharacter{2074}{\ensuremath{^4}}
\DeclareUnicodeCharacter{2075}{\ensuremath{^5}}
\DeclareUnicodeCharacter{2076}{\ensuremath{^6}}
\DeclareUnicodeCharacter{2077}{\ensuremath{^7}}
\DeclareUnicodeCharacter{2078}{\ensuremath{^8}}
\DeclareUnicodeCharacter{2079}{\ensuremath{^9}}
\DeclareUnicodeCharacter{207B}{\ensuremath{^-}}
\DeclareUnicodeCharacter{00B0}{\ensuremath{^{\circ}}}
\DeclareUnicodeCharacter{2032}{\ensuremath{^{\prime}}}
\DeclareUnicodeCharacter{2033}{\ensuremath{^{\prime\prime}}}

% Make the "warning" and "notes" sections use a sans-serif font to
% make them stand out more.
\renewenvironment{notice}[2]{
  \def\py@noticetype{#1}
  \csname py@noticestart@#1\endcsname
  \textsf{\textbf{#2}}
}{\csname py@noticeend@\py@noticetype\endcsname}
"""

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# -- Options for the linkcheck builder ----------------------------------------

# A timeout value, in seconds, for the linkcheck builder
linkcheck_timeout = 60



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from pathlib import Path
from ..io import tile_default_url, tile_default_path


def test_tile_default_url():
    url = tile_default_url(order=9, ipix=54321, file_format='fits')
    assert url == 'Norder9/Dir50000/Npix54321.fits'


def test_tile_default_path():
    path = tile_default_path(order=9, ipix=54321, file_format='fits')
    assert path == Path('Norder9/Dir50000/Npix54321.fits')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Common utility functions used throughout the ``hips`` package.

Note that those are mostly for internal notes, if you use them from
external scripts or packages at all, you should not expect stability
from our side and it might be better if you copy over what you need.
"""



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy.tests.helper import remote_data
from ...utils.testing import make_test_wcs_geometry
from ...tiles import HipsSurveyProperties
from ..ui import make_sky_image

make_sky_image_pars = [
    dict(
        file_format='fits',
        shape=(1000, 2000),
        url='http://alasky.unistra.fr/DSS/DSS2Merged/properties',
        data_1=2213,
        data_2=2296,
        data_sum=8756493140,
        dtype='>i2',
        repr='HipsDrawResult(width=1000, height=2000, channels=2, dtype=>i2, format=fits)',
        precise=False
    ),
    dict(
        file_format='jpg',
        shape=(1000, 2000, 3),
        url='https://raw.githubusercontent.com/hipspy/hips-extra/master/datasets/samples/FermiColor/properties',
        data_1=[133, 117, 121],
        data_2=[137, 116, 114],
        data_sum=828908873,
        dtype='uint8',
        repr='HipsDrawResult(width=1000, height=2000, channels=3, dtype=uint8, format=jpg)',
        precise=False
    ),
    dict(
        file_format='jpg',
        shape=(1000, 2000, 3),
        url='https://raw.githubusercontent.com/hipspy/hips-extra/master/datasets/samples/FermiColor/properties',
        data_1=[139, 114, 115],
        data_2=[142, 113, 110],
        data_sum=825148172,
        dtype='uint8',
        repr='HipsDrawResult(width=1000, height=2000, channels=3, dtype=uint8, format=jpg)',
        precise=True
    ),
    dict(
        file_format='png',
        shape=(1000, 2000, 4),
        url='https://raw.githubusercontent.com/hipspy/hips-extra/master/datasets/samples/AKARI-FIS/properties',
        data_1=[224, 216, 196, 255],
        data_2=[227, 217, 205, 255],
        data_sum=1635622838,
        dtype='uint8',
        repr='HipsDrawResult(width=1000, height=2000, channels=3, dtype=uint8, format=png)',
        precise=False
    ),
]


@remote_data
@pytest.mark.parametrize('pars', make_sky_image_pars)
def test_make_sky_image(tmpdir, pars):
    hips_survey = HipsSurveyProperties.fetch(url=pars['url'])
    geometry = make_test_wcs_geometry()

    fetch_opts = dict(fetch_package='urllib', timeout=30, n_parallel=10)
    result = make_sky_image(geometry=geometry, hips_survey=hips_survey, tile_format=pars['file_format'],
                            precise=pars['precise'], fetch_opts=fetch_opts)

    assert result.image.shape == pars['shape']
    assert result.image.dtype == pars['dtype']
    assert repr(result) == pars['repr']
    assert_allclose(np.sum(result.image, dtype=float), pars['data_sum'])
    assert_allclose(result.image[200, 994], pars['data_1'])
    assert_allclose(result.image[200, 995], pars['data_2'])
    result.write_image(str(tmpdir / 'test.' + pars['file_format']))
    result.plot()
    result.report()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
from astropy.tests.helper import remote_data
from numpy.testing import assert_allclose
from ..fetch import fetch_tiles
from ..survey import HipsSurveyProperties
from ..tile import HipsTileMeta

TILE_FETCH_TEST_CASES = [
    dict(
        tile_indices=[69623, 69627, 69628, 69629, 69630, 69631],
        tile_format='fits',
        order=7,
        url='http://alasky.unistra.fr/DSS/DSS2Merged/properties',
        progress_bar=True,
        data=[2101, 1945, 1828, 1871, 2079, 2336],
        fetch_package='urllib',
    ),
    dict(
        tile_indices=[69623, 69627, 69628, 69629, 69630, 69631],
        tile_format='fits',
        order=7,
        url='http://alasky.unistra.fr/DSS/DSS2Merged/properties',
        progress_bar=True,
        data=[2101, 1945, 1828, 1871, 2079, 2336],
        fetch_package='aiohttp',
    ),
]


def make_tile_metas(hips_survey, pars):
    for healpix_pixel_index in pars['tile_indices']:
        yield HipsTileMeta(
            order=pars['order'],
            ipix=healpix_pixel_index,
            frame=hips_survey.astropy_frame,
            file_format=pars['tile_format'],
        )


@pytest.mark.parametrize('pars', TILE_FETCH_TEST_CASES)
@remote_data
def test_fetch_tiles(pars):
    hips_survey = HipsSurveyProperties.fetch(pars['url'])

    tile_metas = list(make_tile_metas(hips_survey, pars))

    tiles = fetch_tiles(
        tile_metas, hips_survey,
        progress_bar=pars['progress_bar'],
        fetch_package=pars['fetch_package'],
    )

    for idx, val in enumerate(pars['data']):
        assert_allclose(tiles[idx].data[0][5], val)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from collections import namedtuple
from typing import Tuple, Union
import numpy as np
from astropy.coordinates import Angle, SkyCoord
from astropy.io.fits import Header
from astropy.wcs import WCS
from astropy.wcs.utils import pixel_to_skycoord, wcs_to_celestial_frame

__all__ = [
    'WCSGeometry',
]

__doctest_skip__ = [
    'WCSGeometry',
    'WCSGeometry.create',
]

Shape = namedtuple('Shape', ['height', 'width'])
"""Helper for 2-dim image shape, to make it clearer what value is width and height."""


class WCSGeometry:
    """Sky image geometry: WCS and image shape.

    Parameters
    ----------
    wcs : `~astropy.wcs.WCS`
        WCS projection object
    width, height : int
        Width and height of the image in pixels

    Examples
    --------
    To create a ``WCSGeometry``, you can create any `~astropy.wcs.WCS`
    and choose an image shape (number of pixels)::

        from astropy.wcs import WCS
        from hips import WCSGeometry
        wcs = WCS(naxis=2)
        wcs.wcs.ctype[0] = 'GLON-AIT'
        wcs.wcs.ctype[1] = 'GLAT-AIT'
        wcs.wcs.crval[0] = 0
        wcs.wcs.crval[1] = 0
        wcs.wcs.crpix[0] = 1000
        wcs.wcs.crpix[1] = 500
        wcs.wcs.cdelt[0] = -0.01
        wcs.wcs.cdelt[1] = 0.01
        geometry =  WCSGeometry(wcs, width=2000, height=1000)

    See also `WCSGeometry.create` as a simpler (but also not quite as flexible)
    way to generate ``WCS`` and ``WCSGeometry`` objects.
    """
    WCS_ORIGIN_DEFAULT = 0
    """Default WCS transform origin, to be used in all WCS pix <-> world calls."""

    def __init__(self, wcs: WCS, width: int, height: int) -> None:
        self.wcs = wcs
        self.shape = Shape(width=width, height=height)

    def __str__(self):
        return (
            'WCSGeometry data:\n'
            f'WCS: {self.wcs}\n'
            f'Shape: {self.shape}\n'
        )

    @property
    def center_pix(self) -> Tuple[float, float]:
        """Image center in pixel coordinates (tuple of x, y)."""
        x = float(self.shape.width - 1) / 2
        y = float(self.shape.height - 1) / 2
        return x, y

    @property
    def pixel_skycoords(self) -> SkyCoord:
        """Grid of sky coordinates of the image pixels (`~astropy.coordinates.SkyCoord`)."""
        y, x = np.indices(self.shape)
        return self.pix_to_sky(x, y)

    @property
    def center_skycoord(self) -> SkyCoord:
        """Image center in sky coordinates (`~astropy.coordinates.SkyCoord`)."""
        return self.pix_to_sky(*self.center_pix)

    def pix_to_sky(self, x, y) -> SkyCoord:
        """Helper function to convert pix to sky coordinates."""
        return pixel_to_skycoord(x, y, self.wcs, self.WCS_ORIGIN_DEFAULT)

    @classmethod
    def create(cls, skydir: SkyCoord, width: int, height: int, fov: Union[str, Angle],
               coordsys: str = 'icrs', projection: str = 'AIT') -> 'WCSGeometry':
        """Create WCS object for given sky image parameters (`WCSGeometry`).

        Parameters
        ----------
        skydir : `~astropy.coordinates.SkyCoord`
            Sky coordinate of the WCS reference point
        width, height : int
            Width and height of the image in pixels
        fov : str or `~astropy.coordinates.Angle`
            Field of view
        coordsys : {'icrs', 'galactic'}
            Coordinate system
        projection : str
            Projection of the WCS object.
            To see list of supported projections
            visit: http://docs.astropy.org/en/stable/wcs/#supported-projections

        Examples
        --------
        >>> from astropy.coordinates import SkyCoord
        >>> from hips import WCSGeometry
        >>> skycoord = SkyCoord(10, 20, unit='deg')
        >>> geometry = WCSGeometry.create(
        ...     skydir=SkyCoord(0, 0, unit='deg', frame='galactic'),
        ...     width=2000, height=1000, fov='3 deg',
        ...     coordsys='galactic', projection='AIT',
        ... )
        >>> geometry.wcs
        Number of WCS axes: 2
        CTYPE : 'GLON-AIT'  'GLAT-AIT'
        CRVAL : 0.0  0.0
        CRPIX : 500.0  1000.0
        PC1_1 PC1_2  : 1.0  0.0
        PC2_1 PC2_2  : 0.0  1.0
        CDELT : -0.0015  0.0015
        NAXIS : 0  0
        >>> geometry.shape
        Shape(width=2000, height=1000)
        """
        fov = Angle(fov)
        crpix = float(width / 2), float(height / 2)
        cdelt = float(fov.degree) / float(max(width, height))

        w = WCS(naxis=2)

        if coordsys == 'icrs':
            w.wcs.ctype[0] = f'RA---{projection}'
            w.wcs.ctype[1] = f'DEC--{projection}'
            w.wcs.crval[0] = skydir.icrs.ra.deg
            w.wcs.crval[1] = skydir.icrs.dec.deg
        elif coordsys == 'galactic':
            w.wcs.ctype[0] = f'GLON-{projection}'
            w.wcs.ctype[1] = f'GLAT-{projection}'
            w.wcs.crval[0] = skydir.galactic.l.deg
            w.wcs.crval[1] = skydir.galactic.b.deg
        else:
            raise ValueError('Unrecognized coordinate system.')  # pragma: no cover

        w.wcs.crpix[0] = crpix[0]
        w.wcs.crpix[1] = crpix[1]

        w.wcs.cdelt[0] = -cdelt
        w.wcs.cdelt[1] = cdelt

        w = WCS(w.to_header())

        return cls(w, width, height)

    @classmethod
    def create_from_dict(cls, params: dict) -> 'WCSGeometry':
        """Create WCS object from a dictionary (`WCSGeometry`).

        The extra options are passed to the ``create`` class method, it can take the following parameters:

        * target
        * width
        * height
        * fov
        * coordsys
        * projection

        For detailed description, see the ``create`` class method's docstring.
        """
        skycoord = SkyCoord.from_name(params.pop('target'), frame=params['coordsys'])
        return cls.create(skycoord, **params)

    @classmethod
    def make(cls, geometry: Union[dict, 'WCSGeometry']) -> 'WCSGeometry':
        """Convenience constructor for create_from_dict classmethod or existing object (`WCSGeometry`)."""
        if isinstance(geometry, str):
            return WCSGeometry.create_from_dict(geometry)
        elif isinstance(geometry, WCSGeometry):
            return geometry
        else:
            raise TypeError(f'geometry must be of type dict or `WCSGeometry`. You gave {type(geometry)}')

    @property
    def celestial_frame(self) -> str:
        """Celestial frame for the given WCS (str).

        Calls `~astropy.wcs.utils.wcs_to_celestial_frame`.
        """
        return wcs_to_celestial_frame(self.wcs)

    @property
    def fits_header(self) -> Header:
        """FITS header for the given WCS (`~astropy.io.fits.Header`)."""
        return self.wcs.to_header()



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import glob
import os

import ah_bootstrap
from setuptools import setup

# A dirty hack to get around some early import/configurations ambiguities
import builtins

builtins._ASTROPY_SETUP_ = True

from astropy_helpers.setup_helpers import (register_commands, get_debug_option,
                                           get_package_info)
from astropy_helpers.git_helpers import get_git_devstr
from astropy_helpers.version_helpers import generate_version_py

# Get some values from the setup.cfg
from configparser import ConfigParser

conf = ConfigParser()
conf.read(['setup.cfg'])
metadata = dict(conf.items('metadata'))

PACKAGENAME = metadata['package_name']
DESCRIPTION = metadata['description']
AUTHOR = metadata['author']
AUTHOR_EMAIL = metadata['author_email']
LICENSE = metadata['license']
URL = metadata['url']

with open('LONG_DESCRIPTION.rst') as f:
    LONG_DESCRIPTION = f.read()

# Store the package name in a built-in variable so it's easy
# to get from other parts of the setup infrastructure
builtins._ASTROPY_PACKAGE_NAME_ = PACKAGENAME

# VERSION should be PEP440 compatible (http://www.python.org/dev/peps/pep-0440)
VERSION = metadata.get('version')

# Indicates if this version is a release version
RELEASE = 'dev' not in VERSION

if not RELEASE:
    VERSION += get_git_devstr(False)

# Populate the dict of setup command overrides; this should be done before
# invoking any other functionality from distutils since it can potentially
# modify distutils' behavior.
cmdclassd = register_commands(PACKAGENAME, VERSION, RELEASE)

# Freeze build information in version.py
generate_version_py(PACKAGENAME, VERSION, RELEASE,
                    get_debug_option(PACKAGENAME))

# Treat everything in scripts except README* as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
           if not os.path.basename(fname).startswith('README')]

# Get configuration information from all of the various subpackages.
# See the docstring for setup_helpers.update_package_files for more
# details.
package_info = get_package_info()

# Add the project-global data
package_info['package_data'].setdefault(PACKAGENAME, [])
package_info['package_data'][PACKAGENAME].append('data/*')

# Define entry points for command-line scripts
entry_points = {'console_scripts': []}

if conf.has_section('entry_points'):
    entry_point_list = conf.items('entry_points')
    for entry_point in entry_point_list:
        entry_points['console_scripts'].append('{0} = {1}'.format(
            entry_point[0], entry_point[1]))

# Include all .c files, recursively, including those generated by
# Cython, since we can not do this in MANIFEST.in with a "dynamic"
# directory name.
c_files = []
for root, dirs, files in os.walk(PACKAGENAME):
    for filename in files:
        if filename.endswith('.c'):
            c_files.append(
                os.path.join(
                    os.path.relpath(root, PACKAGENAME), filename))
package_info['package_data'][PACKAGENAME].extend(c_files)

# Note that requires and provides should not be included in the call to
# ``setup``, since these are now deprecated. See this link for more details:
# https://groups.google.com/forum/#!topic/astropy-dev/urYO8ckB2uM

install_requires = [
    'numpy>=1.11',
    'astropy>=1.3',
    'astropy-healpix',
    'scikit-image',
    'Pillow',
]

extras_require = dict(
    recommended=[
        'matplotlib>=2.0',
        'reproject>=0.3.1',
        'tqdm',
        'aiohttp',
    ],
    develop=[
        'matplotlib>=2.0',
        'reproject>=0.3.1',
        'pytest>=3.0',
        'mypy>=0.501',
        'tqdm',
        'aiohttp',
    ],
)

classifiers = [
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Scientific/Engineering :: Astronomy',
    'Development Status :: 3 - Alpha',
]

setup(
    name=PACKAGENAME,
    version=VERSION,
    description=DESCRIPTION,
    scripts=scripts,
    install_requires=install_requires,
    extras_require=extras_require,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    url=URL,
    long_description=LONG_DESCRIPTION,
    classifiers=classifiers,
    cmdclass=cmdclassd,
    zip_safe=False,
    use_2to3=False,
    entry_points=entry_points,
    **package_info
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import asyncio
import urllib.request
import concurrent.futures
from typing import List
from ..tiles import HipsSurveyProperties, HipsTile, HipsTileMeta

__all__ = [
    'fetch_tiles',
]

__doctest_skip__ = [
    'fetch_tiles',
]


def fetch_tiles(tile_metas: List[HipsTileMeta], hips_survey: HipsSurveyProperties,
                progress_bar: bool = True, n_parallel: int = 5,
                timeout: float = 10, fetch_package: str = 'urllib') -> List[HipsTile]:
    """Fetch a list of HiPS tiles.

    This function fetches a list of HiPS tiles based
    on their URLs, which are generated using ``hips_survey``
    and ``tile_metas``.

    The tiles are then fetched asynchronously using ``urllib`` or ``aiohttp``.

    Parameters
    ----------
    tile_metas : list
        Python list of `~hips.HipsTileMeta`
    hips_survey : `~hips.HipsSurveyProperties`
        HiPS survey properties
    progress_bar : bool
        Show a progress bar for tile fetching and drawing
    n_parallel : int
        Number of tile fetch web requests to make in parallel
    timeout : float
        Seconds to timeout for fetching a HiPS tile
    fetch_package : {'urllib', 'aiohttp'}
        Package to use for fetching HiPS tiles

    Examples
    --------
    Define a list of tiles we want::

        from hips import HipsSurveyProperties, HipsTileMeta
        from hips import fetch_tiles
        url = 'http://alasky.unistra.fr/DSS/DSS2Merged/properties'
        hips_survey = HipsSurveyProperties.fetch(url)
        tile_indices = [69623, 69627, 69628, 69629, 69630, 69631]
        tile_metas = []
        for healpix_pixel_index in tile_indices:
            tile_meta = HipsTileMeta(
               order=7,
               ipix=healpix_pixel_index,
               frame=hips_survey.astropy_frame,
               file_format='fits',
            )
            tile_metas.append(tile_meta)

    Fetch all tiles (in parallel)::

        tiles = fetch_tiles(tile_metas, hips_survey)

    Returns
    -------
    tiles : list
        A Python list of `~hips.HipsTile`
    """
    if fetch_package == 'aiohttp':
        fetch_fct = tiles_aiohttp
    elif fetch_package == 'urllib':
        fetch_fct = tiles_urllib
    else:
        raise ValueError(f'Invalid package name: {fetch_package}')

    tiles = fetch_fct(tile_metas, hips_survey, progress_bar, n_parallel, timeout)

    # Sort tiles to match the tile_meta list
    # TODO: this doesn't seem like a great solution.
    # Use OrderedDict instead?
    out = []
    for tile_meta in tile_metas:
        for tile in tiles:
            if tile.meta == tile_meta:
                out.append(tile)
                continue
    return out


def fetch_tile_urllib(url: str, meta: HipsTileMeta, timeout: float) -> HipsTile:
    """Fetch a HiPS tile asynchronously."""
    with urllib.request.urlopen(url, timeout=timeout) as conn:
        raw_data = conn.read()
        return HipsTile(meta, raw_data)


def tiles_urllib(tile_metas: List[HipsTileMeta], hips_survey: HipsSurveyProperties,
                 progress_bar: bool, n_parallel, timeout: float) -> List[HipsTile]:
    """Generator function to fetch HiPS tiles from a remote URL."""
    with concurrent.futures.ThreadPoolExecutor(max_workers=n_parallel) as executor:
        futures = []
        for meta in tile_metas:
            url = hips_survey.tile_url(meta)
            future = executor.submit(fetch_tile_urllib, url, meta, timeout)
            futures.append(future)

        futures = concurrent.futures.as_completed(futures)
        if progress_bar:
            from tqdm import tqdm
            futures = tqdm(futures, total=len(tile_metas), desc='Fetching tiles')

        tiles = []
        for future in futures:
            tiles.append(future.result())

    return tiles


async def fetch_tile_aiohttp(url: str, meta: HipsTileMeta, session, timeout: float) -> HipsTile:
    """Fetch a HiPS tile asynchronously using aiohttp."""
    async with session.get(url, timeout=timeout) as response:
        raw_data = await response.read()
        return HipsTile(meta, raw_data)


async def fetch_all_tiles_aiohttp(tile_metas: List[HipsTileMeta], hips_survey: HipsSurveyProperties,
                                  progress_bar: bool, n_parallel: int, timeout: float) -> List[HipsTile]:
    """Generator function to fetch HiPS tiles from a remote URL using aiohttp."""
    import aiohttp

    connector = aiohttp.TCPConnector(limit=n_parallel)
    async with aiohttp.ClientSession(connector=connector) as session:
        futures = []
        for meta in tile_metas:
            url = hips_survey.tile_url(meta)
            future = asyncio.ensure_future(fetch_tile_aiohttp(url, meta, session, timeout))
            futures.append(future)

        futures = asyncio.as_completed(futures)
        if progress_bar:
            from tqdm import tqdm
            futures = tqdm(futures, total=len(tile_metas), desc='Fetching tiles')

        tiles = []
        for future in futures:
            tiles.append(await future)

    return tiles


def tiles_aiohttp(tile_metas: List[HipsTileMeta], hips_survey: HipsSurveyProperties,
                  progress_bar: bool, n_parallel: int, timeout: float) -> List[HipsTile]:
    return asyncio.get_event_loop().run_until_complete(
        fetch_all_tiles_aiohttp(tile_metas, hips_survey, progress_bar, n_parallel, timeout)
    )



"""
This package contains utilities and extensions for the Astropy sphinx
documentation.  In particular, the `astropy.sphinx.conf` should be imported by
the sphinx ``conf.py`` file for affiliated packages that wish to make use of
the Astropy documentation format. Note that some sphinx extensions which are
bundled as-is (numpydoc and sphinx-automodapi) are included in
astropy_helpers.extern rather than astropy_helpers.sphinx.ext.
"""



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""HEALPix and HiPS utility functions."""
from functools import lru_cache
import numpy as np
from astropy_healpix import HEALPix
from astropy_healpix.core import level_to_nside
from astropy.coordinates import SkyCoord, Angle, ICRS, Galactic, BaseCoordinateFrame
from .wcs import WCSGeometry

__all__ = [
    'healpix_order_to_npix',
    'hips_order_for_pixel_resolution',
    'healpix_pixel_corners',

    'healpix_pixels_in_sky_image',
    'hips_tile_healpix_ipix_array',
]

__doctest_skip__ = [
    'healpix_pixel_corners',
    'healpix_pixels_in_sky_image',
]


def healpix_order_to_npix(order: int) -> int:
    """HEALPix order to npix."""
    return HEALPix(nside=2 ** order, order='nested').npix


def healpix_pixel_corners(order: int, ipix: int, frame: str) -> SkyCoord:
    """Returns an array containing the angle (theta and phi) in radians.

    This function calls `healpy.boundaries` to compute the four corners of a HiPS tile.

    It's not documented, but apparently the order of the corners is always as follows:

    1. North (N)
    2. West (W)
    3. South (S)
    4. East (E)

    Parameters
    ----------
    order : int
        HEALPix ``order`` parameter
    ipix : int
        HEALPix pixel index
    frame : {'icrs', 'galactic', 'ecliptic'}
        Sky coordinate frame

    Returns
    -------
    corners : `~astropy.coordinates.SkyCoord`
        Sky coordinates (array of length 4).
    """
    frame = make_frame(frame)
    hp = HEALPix(nside=2 ** order, order='nested', frame=frame)
    return hp.boundaries_skycoord(ipix, step=1)[0]


def healpix_pixels_in_sky_image(geometry: WCSGeometry, order: int, healpix_frame: str) -> np.ndarray:
    """Compute HEALPix pixels within a given sky image.

    The algorithm used is as follows:

    * compute the sky position of every pixel in the image using the given ``geometry``
    * compute the HEALPix pixel index for every pixel using `healpy.pixelfunc.ang2pix`
    * compute the unique set of HEALPix pixel values that occur via `numpy.unique`

    Parameters
    ----------
    geometry : `WCSGeometry`
        Sky image WCS geometry
    order : int
        HEALPix order
    healpix_frame : {'icrs', 'galactic', 'ecliptic'}
        HEALPix coordinate frame

    Returns
    -------
    pixels : `numpy.ndarray`
        HEALPix pixel numbers

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> from hips import WCSGeometry
    >>> from hips.utils.healpix import healpix_pixels_in_sky_image
    >>> skycoord = SkyCoord(10, 20, unit="deg")
    >>> geometry = WCSGeometry.create(
    ...     skydir=skycoord, shape=(10, 20),
    ...     coordsys='CEL', projection='AIT',
    ...     cdelt=1.0, crpix=(1., 1.),
    ... )
    >>> healpix_pixels_in_sky_image(geometry, order=3, healpix_frame='galactic')
    array([321, 611, 614, 615, 617, 618, 619, 620, 621, 622])
    """
    hp = HEALPix(nside=2 ** order, order='nested', frame=healpix_frame)
    skycoord = geometry.pixel_skycoords  # .transform_to(healpix_frame)
    ipix = hp.skycoord_to_healpix(skycoord)
    return np.unique(ipix)


def hips_order_for_pixel_resolution(tile_width: int, resolution: float) -> int:
    """Find the HiPS tile order that will result in a given pixel resolution.

    Parameters
    ----------
    tile_width : int
        HiPS tile width
    resolution : float
        Sky image angular resolution (pixel size in degrees)

    Returns
    -------
    candidate_tile_order : int
        Best HiPS tile order
    """
    resolution = Angle(resolution, unit='deg')
    nside = pixel_resolution_to_nside(resolution * tile_width, round='up')
    return nside_to_level(nside)


@lru_cache(maxsize=None)
def hips_tile_healpix_ipix_array(shift_order: int) -> np.ndarray:
    """

    TODO: write a good description what this is.
    Note that there is already a mention of this in the high-level docs:
    https://github.com/hipspy/hips/blame/71e593ab7e60767be70d9b2b13398016c35db09a/docs/drawing_algo.rst#L103

    Parameters
    ----------
    shift_order : int
        The HiPS tile "shift order", which is related to the tile
        pixel width as follows: ``tile_width = 2 ** shift_order``.
        Supported range of values: 1 to 16

    Returns
    -------
    shift_ipix_array : `~numpy.ndarray`
        2-dimensional array of HEALPix nested order ``ipix`` values
        for the tile pixels. These numbers are relative to the
        HiPS tile HEALPix index, which needs to be added.

    Examples
    --------
    TODO: give examples here, or elsewhere from where this helper is called?
    """
    # Sanity check, prevent users from shooting themselves in the foot here
    # and waste a lot of CPU and memory.
    if not isinstance(shift_order, int):
        raise TypeError('The `shift_order` option must by of type `int`.')
    # Usually tiles have ``shift_order == 9``, i.e. ``tile_width == 512`
    # There's no examples for very high shift order,
    # so the ``shift_oder == 16`` limit (``tile_width == 65536``) here should be OK.
    if shift_order < 1 or shift_order > 16:
        raise ValueError('The `shift_order` must be in the range 1 to 16.')

    if shift_order == 1:
        return np.array([[0, 1], [2, 3]])
    else:
        # Create 4 tiled copies of the parent
        ipix_parent = hips_tile_healpix_ipix_array(shift_order - 1)
        data1 = np.tile(ipix_parent, reps=(2, 2))

        # Add the right offset values to each of the 4 parts
        repeats = 2 ** (shift_order - 1)
        data2 = (repeats ** 2) * np.array([[0, 1], [2, 3]])
        data2 = np.repeat(data2, repeats, axis=0)
        data2 = np.repeat(data2, repeats, axis=1)

        return data1 + data2


# TODO: remove this function and call the one in `astropy_healpix`
# as soon as astropy-healpix v0.3 is released
def pixel_resolution_to_nside(resolution, round='nearest'):
    resolution = Angle(resolution).radian
    pixel_area = resolution * resolution
    npix = 4 * np.pi / pixel_area
    nside = np.sqrt(npix / 12)

    # Now we have to round to the closest ``nside``
    # Since ``nside`` must be a power of two,
    # we first compute the corresponding ``level = log2(nside)`
    # round the level and then go back to nside
    level = np.log2(nside)

    if round == 'up':
        level = np.array(level, dtype=int) + 1
    elif round == 'nearest':
        level = np.array(level + 0.5, dtype=int)
    elif round == 'down':
        level = np.array(level, dtype=int)
    else:
        raise ValueError('Invalid value for round: {!r}'.format(round))

    return level_to_nside(np.clip(level, 0, None))


# TODO: move to astropy-healpix
# Call from above as: HEALPix(nside, order='nested').level
def nside_to_level(nside):
    level = np.log2(nside)
    return np.round(level).astype(int)


# TODO: move to astropy-healpix
# or maybe call this? `astropy.coordinates.frame_transform_graph.lookup_name(frame)`
# See https://github.com/astropy/astropy-healpix/issues/56#issuecomment-336803290
def make_frame(frame):
    if isinstance(frame, BaseCoordinateFrame):
        return frame
    elif isinstance(frame, str):
        if frame == 'icrs':
            return ICRS()
        elif frame == 'galactic':
            return Galactic()
        else:
            raise ValueError(f'Invalid value for frame: {frame!r}')
    else:
        raise TypeError(f'Invalid type for frame: {type(frame)}')



# This module defines functions that can be used to check whether OpenMP is
# available and if so what flags to use. To use this, import the
# add_openmp_flags_if_available function in a setup_package.py file where you
# are defining your extensions:
#
#     from astropy_helpers.openmp_helpers import add_openmp_flags_if_available
#
# then call it with a single extension as the only argument:
#
#     add_openmp_flags_if_available(extension)
#
# this will add the OpenMP flags if available.

from __future__ import absolute_import, print_function

import os
import sys
import glob
import tempfile
import subprocess

from distutils import log
from distutils.ccompiler import new_compiler
from distutils.sysconfig import customize_compiler
from distutils.errors import CompileError, LinkError

from .setup_helpers import get_compiler_option

__all__ = ['add_openmp_flags_if_available']


CCODE = """
#include <omp.h>
#include <stdio.h>
int main(void) {
  #pragma omp parallel
  printf("nthreads=%d\\n", omp_get_num_threads());
  return 0;
}
"""


def add_openmp_flags_if_available(extension):
    """
    Add OpenMP compilation flags, if available (if not a warning will be
    printed to the console and no flags will be added)

    Returns `True` if the flags were added, `False` otherwise.
    """

    ccompiler = new_compiler()
    customize_compiler(ccompiler)

    tmp_dir = tempfile.mkdtemp()

    start_dir = os.path.abspath('.')

    if get_compiler_option() == 'msvc':
        compile_flag = '-openmp'
        link_flag = ''
    else:
        compile_flag = '-fopenmp'
        link_flag = '-fopenmp'

    try:

        os.chdir(tmp_dir)

        with open('test_openmp.c', 'w') as f:
            f.write(CCODE)

        os.mkdir('objects')

        # Compile, link, and run test program
        ccompiler.compile(['test_openmp.c'], output_dir='objects', extra_postargs=[compile_flag])
        ccompiler.link_executable(glob.glob(os.path.join('objects', '*')), 'test_openmp', extra_postargs=[link_flag])
        output = subprocess.check_output('./test_openmp').decode(sys.stdout.encoding or 'utf-8').splitlines()

        if 'nthreads=' in output[0]:
            nthreads = int(output[0].strip().split('=')[1])
            if len(output) == nthreads:
                using_openmp = True
            else:
                log.warn("Unexpected number of lines from output of test OpenMP "
                         "program (output was {0})".format(output))
                using_openmp = False
        else:
            log.warn("Unexpected output from test OpenMP "
                     "program (output was {0})".format(output))
            using_openmp = False

    except (CompileError, LinkError):

        using_openmp = False

    finally:

        os.chdir(start_dir)

    if using_openmp:
        log.info("Compiling Cython extension with OpenMP support")
        extension.extra_compile_args.append(compile_flag)
        extension.extra_link_args.append(link_flag)
    else:
        log.warn("Cannot compile Cython extension with OpenMP, reverting to non-parallel code")

    return using_openmp



# this contains imports plugins that configure py.test for astropy tests.
# by importing them here in conftest.py they are discoverable by py.test
# no matter how it is invoked within the source tree.

from astropy.tests.pytest_plugins import *

## Uncomment the following line to treat all DeprecationWarnings as
## exceptions
# enable_deprecations_as_exceptions()

PYTEST_HEADER_MODULES.clear()
PYTEST_HEADER_MODULES.update([
    ('numpy', 'numpy'),
    ('Pillow', 'PIL'),
    ('scikit-image', 'skimage'),
    ('Astropy', 'astropy'),
    ('astropy-healpix', 'astropy_healpix'),
    ('aiohttp', 'aiohttp'),
    ('reproject', 'reproject'),
    ('matplotlib', 'matplotlib'),
])


# This is to figure out the affiliated package version, rather than
# using Astropy's
try:
    from .version import version
except ImportError:
    version = 'dev'

import os

packagename = os.path.basename(os.path.dirname(__file__))
TESTED_VERSIONS[packagename] = version



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from collections import OrderedDict
from io import StringIO
from csv import DictWriter
import urllib.request
from typing import List, Union
from astropy.table import Table
from .tile import HipsTileMeta

__all__ = [
    'HipsSurveyProperties',
    'HipsSurveyPropertiesList',
]

__doctest_skip__ = [
    'HipsSurveyProperties',
    'HipsSurveyPropertiesList',
]


class HipsSurveyProperties:
    """HiPS properties container.

    Parameters
    ----------
    data : `~collections.OrderedDict`
        HiPS survey properties

    Examples
    --------
    >>> from hips import HipsSurveyProperties
    >>> url = 'http://alasky.unistra.fr/DSS/DSS2Merged/properties'
    >>> hips_survey_property = HipsSurveyProperties.fetch(url)
    >>> hips_survey_property.base_url
    'http://alasky.u-strasbg.fr/DSS/DSS2Merged'
    """
    hips_to_astropy_frame_mapping = OrderedDict([
        ('equatorial', 'icrs'),
        ('galactic', 'galactic'),
        ('ecliptic', 'ecliptic'),
    ])
    """HIPS to Astropy SkyCoord frame string mapping."""

    def __init__(self, data: OrderedDict) -> None:
        self.data = data

    @classmethod
    def from_name(cls, name: str) -> 'HipsSurveyProperties':
        """Create object from Survey ID (`HipsSurveyProperties`)."""
        # TODO: implement some kind of caching for HipsSurveyPropertiesList
        surveys = HipsSurveyPropertiesList.fetch()
        return surveys.from_name(name)

    @classmethod
    def make(cls, hips_survey: Union[str, 'HipsSurveyProperties']) -> 'HipsSurveyProperties':
        """Convenience constructor for from_string classmethod or existing object (`HipsSurveyProperties`)."""
        if isinstance(hips_survey, str):
            return HipsSurveyProperties.from_name(hips_survey)
        elif isinstance(hips_survey, HipsSurveyProperties):
            return hips_survey
        else:
            raise TypeError(f'hips_survey must be of type str or `HipsSurveyProperties`. You gave {type(hips_survey)}')

    @classmethod
    def read(cls, filename: str) -> 'HipsSurveyProperties':
        """Read from HiPS survey description file (`HipsSurveyProperties`).

        Parameters
        ----------
        filename : str
            HiPS properties filename
        """
        with open(filename) as fh:
            text = fh.read()

        return cls.parse(text)

    @classmethod
    def fetch(cls, url: str) -> 'HipsSurveyProperties':
        """Read from HiPS survey description file from remote URL (`HipsSurveyProperties`).

        Parameters
        ----------
        url : str
            URL containing HiPS properties
        """

        with urllib.request.urlopen(url) as response:
            text = response.read().decode('utf-8')
        return cls.parse(text, url)

    @classmethod
    def parse(cls, text: str, url: str = None) -> 'HipsSurveyProperties':
        """Parse HiPS survey description text (`HipsSurveyProperties`).

        Parameters
        ----------
        text : str
            Text containing HiPS survey properties
        url : str
            Properties URL of HiPS
        """
        data: OrderedDict = OrderedDict()
        for line in text.split('\n'):
            # Skip empty or comment lines
            if line == '' or line.startswith('#'):
                continue
            try:
                key, value = [_.strip() for _ in line.split('=')]
                data[key] = value
            except ValueError:
                # Skip bad lines (silently, might not be a good idea to do this)
                continue

        if url is not None:
            data['properties_url'] = url.rsplit('/', 1)[0]

        return cls(data)

    @property
    def title(self) -> str:
        """HiPS title (str)."""
        return self.data['obs_title']

    @property
    def hips_version(self) -> str:
        """HiPS version (str)."""
        return self.data['hips_version']

    @property
    def hips_frame(self) -> str:
        """HiPS coordinate frame (str)."""
        return self.data['hips_frame']

    @property
    def astropy_frame(self) -> str:
        """Astropy coordinate frame (str)."""
        return self.hips_to_astropy_frame_mapping[self.hips_frame]

    @property
    def hips_order(self) -> int:
        """HiPS order (int)."""
        return int(self.data['hips_order'])

    @property
    def tile_width(self) -> int:
        """HiPS tile width"""
        try:
            return int(self.data['hips_tile_width'])
        except KeyError:
            return 512

    @property
    def tile_format(self) -> str:
        """HiPS tile format (str)."""
        return self.data['hips_tile_format']

    @property
    def hips_service_url(self) -> str:
        """HiPS service base URL (str)."""
        return self.data['hips_service_url']

    @property
    def base_url(self) -> str:
        """HiPS access URL"""
        try:
            return self.data['hips_service_url']
        except KeyError:
            try:
                return self.data['moc_access_url'].rsplit('/', 1)[0]
            except KeyError:
                try:
                    return self.data['properties_url']
                except KeyError:
                    raise ValueError('URL does not exist!')

    def tile_url(self, tile_meta: HipsTileMeta) -> str:
        """Tile URL on the server (str)."""
        return self.base_url + '/' + tile_meta.tile_default_url


class HipsSurveyPropertiesList:
    """HiPS survey properties list.

    Parameters
    ----------
    data : list
        Python list of `~hips.HipsSurveyProperties`

    Examples
    --------
    Fetch the list of available HiPS surveys from CDS:

    >>> from hips import HipsSurveyPropertiesList
    >>> surveys = HipsSurveyPropertiesList.fetch()

    Look at the results:

    >>> len(surveys.data)
    335
    >>> survey = surveys.data[0]
    >>> survey.title
    '2MASS H (1.66 microns)'
    >>> survey.hips_order
    9

    You can make a `astropy.table.Table` of available HiPS surveys:

    >>> table = surveys.table

    and then do all the operations that Astropy table supports, e.g.

    >>> table[['ID', 'hips_order', 'hips_service_url']][[1, 30, 42]]
    >>> table.show_in_browser(jsviewer=True)
    >>> table.show_in_notebook()
    >>> table.to_pandas()

    """
    DEFAULT_URL = ('http://alasky.unistra.fr/MocServer/query?'
                   'hips_service_url=*&dataproduct_type=!catalog&dataproduct_type=!cube&get=record')
    """Default URL to fetch HiPS survey list from CDS."""

    def __init__(self, data: List[HipsSurveyProperties]) -> None:
        self.data = data

    @classmethod
    def read(cls, filename: str) -> 'HipsSurveyPropertiesList':
        """Read HiPS list from file (`HipsSurveyPropertiesList`).

        Parameters
        ----------
        filename : str
            HiPS list filename
        """
        with open(filename, encoding='utf-8', errors='ignore') as fh:
            text = fh.read()

        return cls.parse(text)

    @classmethod
    def fetch(cls, url: str = None) -> 'HipsSurveyPropertiesList':
        """Fetch HiPS list text from remote location (`HipsSurveyPropertiesList`).

        Parameters
        ----------
        url : str
            HiPS list URL
        """
        url = url or cls.DEFAULT_URL
        with urllib.request.urlopen(url) as response:
            text = response.read().decode('utf-8', errors='ignore')
        return cls.parse(text)

    @classmethod
    def parse(cls, text: str) -> 'HipsSurveyPropertiesList':
        """Parse HiPS list text (`HipsSurveyPropertiesList`).

        Parameters
        ----------
        text : str
            HiPS list text
        """
        data = []
        for properties_text in text.split('\n\n'):
            properties = HipsSurveyProperties.parse(properties_text)
            data.append(properties)

        return cls(data)

    @property
    def table(self) -> Table:
        """Table with HiPS survey infos (`~astropy.table.Table`)."""
        # There are two aspects that make creating a `Table` from the data difficult:
        # 1. Not all fields are present for the different surveys
        # 2. All data is stored as strings, numbers haven't been converted yet
        #
        # It might not be the best solution, but the following code does the conversion
        # by going via an intermediate CVS string, which Table can parse directly
        rows = [properties.data for properties in self.data]
        fieldnames = sorted({key for row in rows for key in row})
        buffer = StringIO()
        writer = DictWriter(buffer, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(rows)
        return Table.read(buffer.getvalue(), format='ascii.csv', guess=False)

    def from_name(self, name: str) -> 'HipsSurveyProperties':
        """Return a matching HiPS survey (`HipsSurveyProperties`)."""
        for survey in self.data:
            if survey.data['ID'].strip() == name.strip():
                return survey

        raise KeyError(f'Survey not found: {name}')
